package com.unitech.rfidreader.Listener;

import android.annotation.SuppressLint;
import android.app.AlertDialog;
import android.content.Context;
import android.content.Intent;
import android.nfc.Tag;
import android.os.AsyncTask;
import android.text.Html;
import android.util.Log;
import android.view.LayoutInflater;
import android.view.View;
import android.view.View.OnClickListener;
import android.widget.LinearLayout;
import android.widget.TableLayout;
import android.widget.TableRow;
import android.widget.TextView;

import com.thingmagic.ReadExceptionListener;
import com.thingmagic.ReadListener;
import com.thingmagic.Reader;
import com.thingmagic.ReaderException;
import com.thingmagic.SerialReader;
import com.thingmagic.StatusListener;
import com.thingmagic.TagReadData;
//import com.unitech.rfidreader.BuildConfig;
import com.unitech.rfidreader.R;
import com.unitech.rfidreader.ReaderActivity;
import com.unitech.rfidreader.TagRecord;
import com.unitech.util.Beeper;
import com.unitech.util.LoggerUtil;

import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Set;
import java.util.Timer;
import java.util.TimerTask;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.ReentrantLock;

@SuppressLint("StaticFieldLeak")
public class ServiceListener implements View.OnClickListener {
    private final String TAG = "ServiceListener";

    private int redColor = 0xffff0000;
    private int textColor = 0xff000000;

    //    private TableLayout table;
    private LayoutInflater inflater;
    private ArrayList<String> addedEPCRecords = new ArrayList<>();

    private ConcurrentHashMap<String, TagRecord> epcToReadDataMap = new ConcurrentHashMap<>();
    //    private double maxReadRate = 0;
    private long queryStartTime = 0;
    private long queryStopTime = 0;

    private ReaderActivity mReaderActivity;
    private Timer timer = new Timer();

    private int crcCount = 0;

    private ReentrantLock readerLock;
    private ReentrantLock stopLock, switchLock;
    private Condition stopCondition;
    private ExecutorService executor;

    private ReadThread readThread;
    private Beeper beeper;

//    private SectionsStateAdapter sectionsStateAdapter = null;
//    private TabLayout tabLayout = null;
//    private ViewPager2 viewPager = null;

    public ServiceListener(ReaderActivity readerActivity) {
        mReaderActivity = readerActivity;
        findAllViewsById();
        beeper = Beeper.getInstance();

        readerLock = new ReentrantLock();
        switchLock = new ReentrantLock();

        stopLock = new ReentrantLock();
        stopCondition = stopLock.newCondition();

        executor = Executors.newSingleThreadExecutor();
    }

    private void findAllViewsById() {
        inflater = (LayoutInflater) mReaderActivity.getSystemService(Context.LAYOUT_INFLATER_SERVICE);
//        errorMsgView = mReaderActivity.findViewById(R.id.error_messages_view);
        textColor = mReaderActivity.getSearchResultTextColor();
//        table = (TableLayout) mReaderActivity.findViewById(R.id.tablelayout);
    }

    private void readButtonStatus(boolean setRead, int enable) {
        mReaderActivity.sendBroadcast(new Intent("ACTION_LOCATE_READING")
                .putExtra("start", setRead)
                .putExtra("enable", enable)
        );

        if (setRead) {
            mReaderActivity.setReadBtnText(mReaderActivity.getString(R.string.stop_reading));
        } else {
            mReaderActivity.setReadBtnText(mReaderActivity.getString(R.string.start_reading));
        }

        switch (enable) {
            case 0:
                mReaderActivity.readBtnEdit(false);
                break;
            case 1:
                mReaderActivity.readBtnEdit(true);
                break;
        }
    }

    private void tagOptionButtonStatus(boolean enable) {
        mReaderActivity.tagOpBtnEdit(enable);
    }

    @Override
    public void onClick(View arg0) {
        try {
            if (mReaderActivity.getReadBtnText().equals(mReaderActivity.getResources().getString(R.string.stop_reading))) {
                stopReading();
            } else if (mReaderActivity.getReadBtnText().equals(mReaderActivity.getResources().getString(R.string.start_reading))) {
                startReading();
            }
        } catch (Exception ex) {
            LoggerUtil.error(TAG, "Exception", ex);
        }
    }

    public void startReading() {
//        if (isRunning()) {
//            LoggerUtil.debug(TAG, isReading() ? "It is reading" : "It is stoping");
//            return;
//        }

        LoggerUtil.debug(TAG, "Trigger start reading");
        switchReading(true);
    }

    public void stopReading() {
//        if (!isReading()) {
//            LoggerUtil.debug(TAG, "It is not reading, so do not need stop");
//            return;
//        }

        LoggerUtil.debug(TAG, "Trigger stop reading");
        switchReading(false);
    }

    private void switchReading(boolean enable) {
        Log.d(TAG, "Enter switchReading()");

//        if (!switchLock.tryLock()) {
//            LoggerUtil.debug(TAG, "Switch lock is been locked");
//            return;
//        }
        switchLock.lock();

        Log.d(TAG, "After switchReading() lock to " + enable);

        if (enable) {
            if (isReading()) {
                LoggerUtil.warn(TAG, "Read thread is still reading, do not need read");
                return;
            }
//            mReaderActivity.setReadBtnText(mReaderActivity.getResources().getString(R.string.stop_reading));
            readButtonStatus(true, -1);

            clearTagRecords();

            readThread = new ReadThread(mReaderActivity, "asyncRead");
            readThread.execute();

            while (!isReading()) {
                Log.d(TAG, "Waiting reading...");
                try {
                    Thread.sleep(10);
                } catch (InterruptedException e) {
                    e.printStackTrace();
                }
            }
            switchLock.unlock();
        } else {// if (!enable && !stopLock.isLocked()) {

            if (isReading()) {
                readThread.setReading(false);
                mReaderActivity.runOnUiThread(new Runnable() {
                    @Override
                    public void run() {
                        readButtonStatus(true, 0);
                        mReaderActivity.setReadBtnText(mReaderActivity.getResources().getString(R.string.stopping));
                    }
                });
                executor.submit(runnableStopReading);
            }


        }

//        switchLock.unlock();
        Log.d(TAG, "Exit switchReading()");
    }

    private Runnable runnableStopReading = new Runnable() {
        @Override
        public void run() {
//            if (isReading() && stopLock.tryLock()) {
            try {
                LoggerUtil.debug(TAG, "runnableStopReading()");
                readThread.setReading(false);
                mReaderActivity.runOnUiThread(new Runnable() {
                    @Override
                    public void run() {
                        readButtonStatus(true, 0);
                        mReaderActivity.setReadBtnText(mReaderActivity.getResources().getString(R.string.stopping));
                    }
                });

                while (isRunning()) {
                    Log.d(TAG, "Waiting stoping...");
                    Thread.sleep(10);
                }

                switchLock.unlock();
//                    stopCondition.await(5, TimeUnit.SECONDS);
            } catch (InterruptedException e) {
                e.printStackTrace();
            }
//                stopLock.unlock();
//            }
        }
    };

    public boolean isReading() {
        if (readThread != null) {
            return readThread.isReading();
        } else {
            return false;
        }
    }

    public boolean isRunning() {
        if (readThread != null) {
            return readThread.isRunning();
        } else {
            return false;
        }
    }

    public String listTags() {
//        String tags = "";
//
//        int index = 1;
//        for (TagRecord tag : new ArrayList<>(epcToReadDataMap.values())) {
//            tags += String.format("%s,%s,%s,%s\r\n", index++, tag.getEpcString(), tag.getRssi(), tag.readCount);
//        }

        return mReaderActivity.tagListAdapter.toString();
    }
    //---------------------------------------------------------------------------------------------

    public void clearTagRecords() {
        addedEPCRecords.clear();
        epcToReadDataMap.clear();
//        table.removeAllViews();
        mReaderActivity.resetTagTable();
        mReaderActivity.setSearchResultCountText("");
        mReaderActivity.setTotalTagCountText("");
        mReaderActivity.setAvgReadText("");
        mReaderActivity.setReadingTimeText("");
        mReaderActivity.setTemperatureText("");
//        mReaderActivity.setMaxReadRate("");
//        maxReadRate = 0;
        queryStartTime = System.currentTimeMillis();
    }

    @SuppressLint("StaticFieldLeak")
    public class ReadThread extends AsyncTask<Void, Integer, ConcurrentHashMap<String, TagRecord>> {
        private final String TAG = "ReadThread";
        private boolean _reading = false, _running = false;
//        private ReaderActivity mReaderActivity;

        private final DecimalFormat df = new DecimalFormat("#.##");

        private String temp_str = null, avg_str = null;
        private String operation;
        private boolean exceptionOccur = false;
        private List<String> exception;

        private Reader _reader;
        private int Temperature = 0;

        private int readLoop = 0;

        ReadExceptionListener exceptionListener = new TagReadExceptionReceiver();
        ReadListener readListener = new PrintListener();
        StatusListener statusListener = new ReaderStatusListener();

        ReadThread(ReaderActivity activity, String operation) {
            this.operation = operation;
//            mReaderActivity = activity;
            _reader = mReaderActivity.reader;
            exception = new ArrayList<>();
        }

        @Override
        protected void onPreExecute() {
            setRunning(true);
            clearTagRecords();
            exception.clear();

            mReaderActivity.connectBtnEdit(false);

//            readButtonStatus(false, 0);
            mReaderActivity.setReadBtnText(mReaderActivity.getResources().getString(R.string.starting));

            mReaderActivity.disableBarcode();
            mReaderActivity.tagOpBtnEdit(false);
            mReaderActivity.saveTagBtnEdit(false);
            mReaderActivity.beepCheckBoxEdit(false);

            mReaderActivity.setSearchResultTextColor(textColor);
            mReaderActivity.setSearchResultCountText(mReaderActivity.getResources().getString(R.string.reading_tags));

            addedEPCRecords = new ArrayList<>();
            epcToReadDataMap = new ConcurrentHashMap<>();
            exceptionOccur = false;
        }

        @Override
        protected ConcurrentHashMap<String, TagRecord> doInBackground(Void... params) {
            try {
                readerLock.lock();
                mReaderActivity.runOnUiThread(new Runnable() {
                    @Override
                    public void run() {
                        readButtonStatus(true, 1);
                    }
                });

                _reader.addReadExceptionListener(exceptionListener);
                _reader.addReadListener(readListener);
                _reader.addStatusListener(statusListener);

                _reader.startReading();

                readerLock.unlock();

                setReading(true);

                queryStartTime = System.currentTimeMillis();
                queryStopTime = queryStartTime;
                refreshReadRate();
                while (isReading()) {
                    /* Waiting till stop reading button is pressed */

//                    generateRandomTag(10);

                    Thread.sleep(5);
                }
                LoggerUtil.debug(TAG, "Reading end...");
            } catch (Exception ex) {
                exception.add(ex.getMessage());
                exceptionOccur = true;
                LoggerUtil.error(TAG, "Exception while reading :", ex);
            } finally {
                LoggerUtil.debug(TAG, "ReadThread stopping");
                timer.cancel();

                _reader.removeReadListener(readListener);
                _reader.removeReadExceptionListener(exceptionListener);
                _reader.removeStatusListener(statusListener);
                if (!exceptionOccur) {
                    _reader.stopReading();
                }

                queryStopTime = System.currentTimeMillis();
                LoggerUtil.debug(TAG, "ReadThread stopped");
            }

            return epcToReadDataMap;
        }

        /**
         * Generate random tags for test
         * @param number The number of tags to generate
         */
        void generateRandomTag(int number) {
            for (int i = 0; i < number; i++) {
                int nTag = (int) (Math.random() * 1000000);
                int nRSSI = (int)(Math.random() * -100);
                int nCount = (int)(1 + (Math.random() * 10));
                String epcString = String.format("00001111222233334%07d", nTag);

                if (epcToReadDataMap.containsKey(epcString)) {
                    TagRecord tempTR = epcToReadDataMap.get(epcString);
                    tempTR.setRssi(nRSSI);
                    tempTR.readCount += nCount;
                } else {
                    TagRecord tagRecord = new TagRecord();
                    tagRecord.setEpcString(epcString);
                    tagRecord.setReadCount(nCount);
                    tagRecord.setRssi(nRSSI);
                    epcToReadDataMap.put(epcString, tagRecord);
                }
            }
        }

        class PrintListener implements ReadListener {
            public void tagRead(Reader r, final TagReadData tr) {
                parseTag(tr);
//                readThread.parseTag(tr);
            }
        }

        class ReaderStatusListener implements StatusListener {
            @Override
            public void statusMessage(Reader reader, SerialReader.StatusReport[] statusReports) {
                if (reader == _reader) {
                    for (int i = 0; i < statusReports.length; i++) {
                        if (statusReports[i] instanceof SerialReader.TemperatureStatusReport) {
                            LoggerUtil.debug(TAG, statusReports[i].toString());
                            Temperature = ((SerialReader.TemperatureStatusReport) statusReports[i]).getTemperature();

                            if (Temperature > 70) {
                                final String ex = "The module temperature rises above 70°C and App stops reading to protect module. ";
                                LoggerUtil.warn(TAG, String.format("Module exception : %s", ex));
//                                readThread.setReading(false);
                                setReading(false);
                                mReaderActivity.runOnUiThread(new Runnable() {
                                    @Override
                                    public void run() {
                                        AlertDialog deleteAlert = new AlertDialog.Builder(mReaderActivity).create();
                                        deleteAlert.setTitle("Module Overheat");
                                        deleteAlert.setMessage(ex);
                                        deleteAlert.show();
                                    }
                                });
                            }

                        } else if (statusReports[i] instanceof SerialReader.FrequencyStatusReport) {
                            LoggerUtil.debug(TAG, statusReports[i].toString());
                        }
                    }
                }
            }
        }

        class TagReadExceptionReceiver implements ReadExceptionListener {
            public void tagReadException(Reader r, ReaderException re) {
                if (re.getMessage().contains("The module has detected high return loss")
                        || re.getMessage().contains("Tag ID buffer full")
                        || re.getMessage().contains("No tags found")) {
                    // exception = "No connected antennas found";
                    /* Continue reading */
                    LoggerUtil.warn(TAG, "Waring exception: ", re);
                } else {
                    LoggerUtil.error(TAG, "Reader exception : ", re);
                    exception.add(re.getMessage());
                    exceptionOccur = true;
                    ///readThread.setReading(false);
                    setReading(false);
                    readThread.publishProgress(-1);
                }
            }
        }

        private void refreshReadRate() {
            timer = new Timer();
            timer.schedule(new TimerTask() {
                @Override
                public void run() {
                    publishProgress(0);
                }
            }, 100, 100);
        }

        private void parseTag(TagReadData tr) {

            String epcString = tr.getTag().epcString();
//            Log.d(TAG, String.format("parseTag: %d read count, epc='%s'", tr.getReadCount(), tr.getTag().epcString()));
            if (mReaderActivity.mActionView.getVisibility() != View.VISIBLE) {
                if (tr.getReadCount() > 0) {
                    beeper.beep();
                }

                if (epcToReadDataMap.keySet().contains(epcString)) {
                    TagRecord tempTR = epcToReadDataMap.get(epcString);
                    tempTR.setRssi(tr.getRssi());
                    tempTR.readCount += tr.getReadCount();
                } else {
                    TagRecord tagRecord = new TagRecord();
                    tagRecord.setEpcString(epcString);
                    tagRecord.setReadCount(tr.getReadCount());
                    tagRecord.setRssi(tr.getRssi());
                    epcToReadDataMap.put(epcString, tagRecord);
                }

            } else {

                if (mReaderActivity.ActionTabPosition == 4) {
                    mReaderActivity.sendBroadcast(new Intent("ACTION_LOCATE_DATA")
                            .putExtra("LOCATE_EPC", epcString)
                            .putExtra("LOCATE_RSSI", tr.getRssi()));
                }
            }
        }

        @Override
        protected void onProgressUpdate(Integer... progress) {
            int progressToken = progress[0];
            if (progressToken == -1) {
                mReaderActivity.setErrorMsgTextColor(redColor);
                StringBuilder sb = new StringBuilder("ERROR :");
                for (String exMsg : exception) {
                    sb.append("\n");
                    sb.append(exMsg);
                }
                temp_str = sb.toString();
                mReaderActivity.showErrorMsg(true);
                mReaderActivity.setErrorMsg(temp_str);

                if (++readLoop >= 100 || !isReading()) {
                    readLoop = 0;
                }
            } else {
//                populateSearchResult(epcToReadDataMap);
                mReaderActivity.UpdateTagList(epcToReadDataMap);
                epcToReadDataMap.clear();

                if (!exceptionOccur) {
                    mReaderActivity.setSearchResultTextColor(textColor);

                    if (crcCount == 0) {
                        mReaderActivity.setSearchResultCountText(Html.fromHtml("<b>Unique Tags :</b> " + mReaderActivity.tagListAdapter.size()));
                    } else {
                        mReaderActivity.setSearchResultCountText(Html.fromHtml("<b>CRC : " + crcCount + " Unique Tags :</b> " + mReaderActivity.tagListAdapter.size()));
                    }

                    mReaderActivity.setTotalTagCountText(Html.fromHtml("<b>Total Tags  :</b> " + mReaderActivity.tagListAdapter.getTotalTags()));

                    long elapsedTime = System.currentTimeMillis() - queryStartTime;
                    double timeTaken = ((double) elapsedTime / 1000);
                    temp_str = df.format(timeTaken) + " sec";

                    mReaderActivity.setReadingTimeText(Html.fromHtml("<b>Time:</b> " + temp_str));

                    double tmpAvg = mReaderActivity.tagListAdapter.getTotalTags() / timeTaken;

//                    if (timeTaken > 1) {
//                        if (tmpAvg > maxReadRate)
//                            maxReadRate = tmpAvg;
//
//                        String strMaxReadRate = df.format(maxReadRate);
//                        mReaderActivity.setMaxReadRate(Html.fromHtml("<b>Max. Read Rate:</b> " + strMaxReadRate));
//                    }

                    avg_str = df.format(tmpAvg);
                    mReaderActivity.setAvgReadText(Html.fromHtml("<b>Avg. Reads Rate:</b> " + avg_str));

                    // Update temperature
                    mReaderActivity.setTemperatureText(Html.fromHtml("<b>Temperature:</b> " + Temperature));

                    if (++readLoop >= 100) {
                        LoggerUtil.info(TAG, String.format("Unique Tags: %d, Total Tags: %d, Time: %s, Avg. Reads Rate: %s",
                                mReaderActivity.tagListAdapter.size(),
                                mReaderActivity.tagListAdapter.getTotalTags(),
                                temp_str,
                                avg_str
                        ));
                        readLoop = 0;
                    }
                }
            }
        }

        @Override
        protected void onPostExecute(ConcurrentHashMap<String, TagRecord> tagList) {
            LoggerUtil.debug(TAG, "Enter onPostExecute()");
            timer.cancel();
            String temp_str;
            if (exceptionOccur) {
                crcCount++;
                mReaderActivity.setErrorMsgTextColor(redColor);
                StringBuilder sb = new StringBuilder("ERROR :");
                for (String exMsg : exception) {
                    sb.append("\n");
                    sb.append(exMsg);
                }
                temp_str = sb.toString();
                mReaderActivity.setErrorMsg(temp_str);
                mReaderActivity.setTotalTagCountText("");
                if (mReaderActivity.tagListAdapter.getTotalTags() > 0 && !operation.equalsIgnoreCase("syncRead")) {
                    mReaderActivity.setTotalTagCountText(Html.fromHtml("<b>Total Tags  :</b> " + mReaderActivity.tagListAdapter.getTotalTags()));
                }
            } else {
                mReaderActivity.UpdateTagList(epcToReadDataMap);
                epcToReadDataMap.clear();

                if (crcCount == 0) {
                    mReaderActivity.setSearchResultCountText(Html.fromHtml("<b>Unique Tags :</b> " + mReaderActivity.tagListAdapter.size()));
                } else {
                    mReaderActivity.setSearchResultCountText(Html.fromHtml("<b>CRC : " + crcCount + " Unique Tags :</b> " + mReaderActivity.tagListAdapter.size()));
                }

                mReaderActivity.setTotalTagCountText(Html.fromHtml("<b>Total Tags  :</b> " + mReaderActivity.tagListAdapter.getTotalTags()));

                long elapsedTime = queryStopTime - queryStartTime;
                double timeTaken = ((double) elapsedTime / 1000);
                temp_str = df.format(timeTaken) + " sec";
                mReaderActivity.setReadingTimeText(Html.fromHtml("<b>Time:</b> " + temp_str));
                mReaderActivity.setAvgReadText(Html.fromHtml("<b>Avg. Reads Rate:</b> " + df.format(mReaderActivity.tagListAdapter.getTotalTags() / timeTaken)));

                LoggerUtil.info(TAG, String.format("Unique Tags: %d, Total Tags: %d, Time: %s, Avg. Reads Rate: %s",
                        mReaderActivity.tagListAdapter.size(),
                        mReaderActivity.tagListAdapter.getTotalTags(),
                        temp_str,
                        df.format(mReaderActivity.tagListAdapter.getTotalTags() / timeTaken)
                ));
            }

            if (exceptionOccur && !exception.isEmpty() && !exception.get(exception.size() - 1).equalsIgnoreCase("No connected antennas found")) {
                disconnectReader();
                mReaderActivity.enableBarcode();
                mReaderActivity.beepCheckBoxEdit(true);
            } else {
                beeper.stopAllBeep();
                mReaderActivity.enableBarcode();
                mReaderActivity.beepCheckBoxEdit(true);

                readButtonStatus(false, 1);
                tagOptionButtonStatus(true);
                mReaderActivity.saveTagBtnEdit(true);

                mReaderActivity.connectBtnEdit(true);
            }
//            stopLock.lock();
//            stopCondition.signalAll();
//            stopLock.unlock();
            setRunning(false);
        }

        private void disconnectReader() {
            mReaderActivity.setConnectBtnText("Connect");
            mReaderActivity.connectBtnEdit(true);

            readButtonStatus(false, 0);
            tagOptionButtonStatus(false);
            mReaderActivity.saveTagBtnEdit(false);
            mReaderActivity.beepCheckBoxEdit(true);

            mReaderActivity.showLayoutFirmware(false);
            mReaderActivity.showFirmwareHint(true);

            mReaderActivity.reader.destroy();
            mReaderActivity.reader = null;

            mReaderActivity.RFIDModulePowerCtrl(false);
            mReaderActivity.RFIDModulePowerCtrl(true);

            if (!exceptionOccur) {
                mReaderActivity.setSearchResultCountText("");
                mReaderActivity.setAvgReadText("");
                mReaderActivity.setReadingTimeText("");
                mReaderActivity.setTemperatureText("");
//                mReaderActivity.setMaxReadRate("");
            }
            beeper.stopAllBeep();
        }

        public boolean isReading() {
            return _reading;
        }

        private void setReading(boolean reading) {
            Log.d(TAG, "ReadThread set reading : " + reading);

            _reading = reading;

            mReaderActivity.runOnUiThread(new Runnable() {
                @Override
                public void run() {
                    //myTabLayoutMediator.isRunning(reading);
                    mReaderActivity.viewPager.setUserInputEnabled(!_reading);
                    LinearLayout tabStrip = ((LinearLayout) mReaderActivity.tabLayout.getChildAt(0));
//                    if (mReaderActivity.serviceListener.isReading()) {

                    tabStrip.setEnabled(!_reading);
                    for (int i = 0; i < tabStrip.getChildCount(); i++) {
                        tabStrip.getChildAt(i).setClickable(!_reading);
                    }
                }
            });
        }

        public boolean isRunning() {
            Log.d(TAG, "ReadThread running is " + _running);
            return _running;
        }

        private synchronized void setRunning(boolean r) {
            Log.d(TAG, "ReadThread set running : " + r);
            _running = r;
        }
    }
}