package com.unitech.rfidreader;

import android.Manifest;
import android.annotation.SuppressLint;
import android.app.ActivityManager;
import android.app.ProgressDialog;
import android.content.BroadcastReceiver;
import android.content.ComponentName;
import android.content.Context;
import android.content.Intent;
import android.content.IntentFilter;
import android.content.pm.PackageInfo;
import android.content.pm.PackageManager;
import android.content.res.Configuration;
import android.graphics.Color;
import android.os.AsyncTask;
import android.os.Bundle;
import android.text.InputType;
import android.text.Spanned;
import android.util.DisplayMetrics;
import android.util.Log;
import android.view.KeyEvent;
import android.view.LayoutInflater;
import android.view.View;
import android.view.View.OnClickListener;
import android.view.WindowManager;
import android.widget.AdapterView;
import android.widget.ArrayAdapter;
import android.widget.Button;
import android.widget.CheckBox;
import android.widget.CompoundButton;
import android.widget.EditText;
import android.widget.ExpandableListView;
import android.widget.LinearLayout;
import android.widget.Spinner;
import android.widget.Switch;
import android.widget.TextView;
import android.widget.Toast;

import com.google.android.material.tabs.TabLayout;
import com.google.android.material.tabs.TabLayoutMediator;
import com.thingmagic.Gen2;
import com.thingmagic.Reader;
import com.thingmagic.SimpleReadPlan;
import com.thingmagic.ReaderException;
import com.thingmagic.TMConstants;
import com.thingmagic.TagData;
import com.thingmagic.TagFilter;
import com.thingmagic.TagProtocol;
import com.unitech.ReaderConnect;

import com.unitech.api.Key.KeySetting;
import com.unitech.api.pogo.PogoCtrl;
import com.unitech.rfidreader.Listener.ConnectionListener;
import com.unitech.rfidreader.Listener.FirmwareListener;
import com.unitech.rfidreader.Listener.SaveTagListener;
import com.unitech.rfidreader.Listener.ServiceListener;
import com.unitech.rfidreader.action.PlaceholderFragment;
import com.unitech.rfidreader.action.SectionsStateAdapter;
import com.unitech.rfidreader.customViews.ExpandableListAdapter;
import com.unitech.rfidreader.customViews.TagListAdapter;
import com.unitech.util.Beeper;
import com.unitech.util.GlobalExceptionHandler;
import com.unitech.util.LoggerUtil;
import com.unitech.util.RFIDConfig;
import com.unitech.util.Utilities;

import java.io.File;

import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Locale;
import java.util.concurrent.ConcurrentHashMap;

import androidx.annotation.NonNull;
import androidx.appcompat.app.AppCompatActivity;
import androidx.core.app.ActivityCompat;
import androidx.localbroadcastmanager.content.LocalBroadcastManager;
import androidx.recyclerview.widget.DividerItemDecoration;
import androidx.recyclerview.widget.LinearLayoutManager;
import androidx.recyclerview.widget.RecyclerView;
import androidx.viewpager2.widget.ViewPager2;
//import androidx.viewpager2.widget.ViewPager2;

public class ReaderActivity extends AppCompatActivity {

    private final String TAG = "ReaderActivity";
    private int KEYCODE_SCAN_EXTENDED = 290;

    //region Spinner Index
    private final int defaultMode = 2;
    private final int Custom_Mode = 3;

    private final int BLF_250K = 0;
    private final int BLF_640K = 1;

    private final int TARI_25 = 0;
    private final int TARI_12_5 = 1;
    private final int TARI_6_25 = 2;

    private final int Session_S0 = 0;
    private final int Session_S1 = 1;
    private final int Session_S2 = 2;
    private final int Session_S3 = 3;

    private final int Encoding_FM0 = 0;
    private final int Encoding_M2 = 1;
    private final int Encoding_M4 = 2;
    private final int Encoding_M8 = 3;

    private final int Target_AB = 0;
    private final int Target_BA = 1;
    private final int Target_A = 2;
    private final int Target_B = 3;
    //endregion

    private final int AntennaOnTimeMax = 1000;
    private final int AntennaOffTimeMin = 100;

    //region Button
    private Button connectButton = null;
    private Button scanButton = null;
    private Button clearButton = null;
    private Button settingsButton = null;
    private Button homeButton = null;
    private Button homeButton1 = null;
    private Button tagOperationButton = null;
    private Button savetagButton = null;
    //endregion

    private DisplayMetrics displaymetrics = new DisplayMetrics();

    //region TextView
    private TextView searchResultCount = null;
    private TextView tableNumberView = null;
    private TextView tableEPCView = null;
    private TextView tableRSSIView = null;
    private TextView tableCountView = null;

    private TextView temperatureView = null;
    private TextView avgReadView = null;
    private TextView totalTagCountView = null;
    private TextView readingTimeView = null;
    private TextView errorMsgView = null;
//    private TextView maxReadRateView = null;

    private TextView appVersionView = null;
    //endregion

    //region Spinner
    private Spinner regionList = null;
    //endregion

    //region Reader Option
    private Spinner baudrateList = null;
    private Spinner operationModeList = null;
    private Spinner powerList = null;

    private EditText readTimeoutView = null;
    private EditText antennaOnTimeView = null;
    private EditText antennaOffTimeView = null;

    private Spinner blfList = null;
    private Spinner tariList = null;
    private Spinner sessionList = null;
    private Spinner encodingList = null;
    private Spinner targetList = null;
    private Spinner qList = null;
    private Switch switchFastSearch = null;
    //endregion

    //region Firmware
    private TextView softwareVersionView = null;
    private TextView hardwareVersionView = null;
    private TextView modelVersionView = null;
    private TextView productGroupVersionView = null;
    private TextView serialNumberView = null;
    private LinearLayout firmware_hint = null;
    private LinearLayout layout_firmware = null;
    //endregion

    //    private TableLayout tagTable = null;
    private RecyclerView tagListView = null;
    public TagListAdapter tagListAdapter = null;

    private int windowWidth;
    public int tableNumberWidth = 0;
    public int tableEPCWidth = 0;
    public int tableRSSIWidth = 0;
    public int tableCountWidth = 0;

    public View readOptions;
    public View firmware;
    public View about;
    public View mActionView;

    private LinearLayout read_result_layout = null;
//    private View tableTags = null;

    public com.unitech.rfidreader.customViews.ExpandableListAdapter expandableListAdapter;

    private CheckBox checkBeep = null;
    public CheckBox enableBarcode = null;
    public EditText barcodeText = null;
    /**
     * The flag indicating whether content is loaded (text is shown) or not (setting view is
     * shown).
     */

    private View mSettingView;
    private View mDisplayView;

    public Reader reader = null;

    public ServiceListener serviceListener;
    private FirmwareListener firmwareListener;

    public Beeper mBeeper;

    private boolean onPauseTrigger = false;

    private boolean targetGunForRFID = true;

    private int skipCount = 0;

    private RFIDConfig properties;

    private boolean firstExecute = true;
    private boolean defaultKeyRemapEnable = true;

    /**
     * Use to save the status before change to barcode
     */
    private boolean preRFIDStatus = false;

    //region Tag Operation
    public SectionsStateAdapter sectionsStateAdapter = null;
    public TabLayout tabLayout = null;
    public ViewPager2 viewPager = null;
    public int ActionTabPosition = 0;
    private boolean initTagOperationFlag = false;

    public String targetEPC = "";

    private ProgressDialog keyRemapDialog;
    SetGunKeyViaResumePauseThread remapThread = null;
    //endregion

    private String FILE_FOLDER = "";
    private String FILE_PATH_KEYS_CONFIG = "";
    private KeySetting keySetting;

    public static final String Action_SelectEPC = "com.unitech.rfidreader.selectEPC";

    //==============================================================================================
    @Override
    public void onCreate(Bundle savedInstanceState) {
        try {
            super.onCreate(savedInstanceState);
            setContentView(R.layout.activity_reader);
            getWindow().setFlags(WindowManager.LayoutParams.FLAG_KEEP_SCREEN_ON, WindowManager.LayoutParams.FLAG_KEEP_SCREEN_ON);
            //--------------------------------------------------------------------------------------
            //Check permission
            if (checkSelfPermission(Manifest.permission.WRITE_EXTERNAL_STORAGE) != PackageManager.PERMISSION_GRANTED) {
                ActivityCompat.requestPermissions(ReaderActivity.this, new String[]{Manifest.permission.WRITE_EXTERNAL_STORAGE}, 1);
            }
            //--------------------------------------------------------------------------------------
            LoggerUtil.initialize(this);
            String path = String.format("%s/%s", this.getExternalFilesDir(null).getAbsolutePath(), LoggerUtil.DIR_LOG);
            File file = new File(path);
            if (!file.exists()) {
                if (file.mkdir()) {
                    LoggerUtil.debug(TAG, "Build AutoTestLog File. ");
                }
            }
            System.out.println(TAG);
            //--------------------------------------------------------------------------------------
            LoggerUtil.debug(TAG, "RFIDReader Version : " + getVersionName(this));

            mSettingView = findViewById(R.id.settings);
            mDisplayView = findViewById(R.id.displayView);
            mActionView = findViewById(R.id.actionView);

            // Initially hide the setting view.
            mSettingView.setVisibility(View.GONE);
            mActionView.setVisibility(View.GONE);

            keyRemapDialog = new ProgressDialog(this);
            keyRemapDialog.setCancelable(false);
            keyRemapDialog.setProgressStyle(ProgressDialog.STYLE_SPINNER);

            keyRemapDialog.setMessage("Switching gun key");
            remapThread = new SetGunKeyViaResumePauseThread(this);

            FILE_FOLDER = String.format("%s/", this.getApplicationContext().getExternalFilesDir(null).getAbsolutePath());
            FILE_PATH_KEYS_CONFIG = String.format("%skeys_config.txt", FILE_FOLDER);

            ReaderActivity activity = this;
            try {
                Utilities.loadLog4j(this);
            } catch (Exception ex) {
                LoggerUtil.error(TAG, "Error loading log4j", ex);
            }
            this.findAllViewsById();
            windowWidth = displaymetrics.widthPixels;

            properties = new RFIDConfig(this.getApplicationContext());

            loadSettingContentView();

            this.findReadOptionAllViewsById();
            this.findFirmwareAllViewsById();

            tagOperationButton.setOnClickListener(new OnClickListener() {
                @Override
                public void onClick(View v) {
                    if (getConnectBtnText().equals(getString(R.string.Disconnect))) {
                        sectionsStateAdapter.SetEPCString("");
                        sectionsStateAdapter.notifyDataSetChanged();
                        showSettingsOrTagsView(2);
                    }
                }
            });

            appVersionView = about.findViewById(R.id.txt_app_version);
            appVersionView.setText(getVersionName(this));

            connectButton.setOnClickListener(new ConnectionListener(this));
            clearButton.setOnClickListener(clearListener);
            savetagButton.setOnClickListener(new SaveTagListener(this));

            serviceListener = new ServiceListener(this);
            scanButton.setOnClickListener(serviceListener);

            settingsButton.setOnClickListener(new OnClickListener() {
                @Override
                public void onClick(View view) {
                    showSettingsOrTagsView(1);
                }
            });

            ArrayAdapter<CharSequence> baudAdapter = ArrayAdapter.createFromResource(this,
                    R.array.baudrate_array, android.R.layout.simple_spinner_item);
            baudAdapter.setDropDownViewResource(android.R.layout.simple_spinner_dropdown_item);
            baudrateList.setAdapter(baudAdapter);
            baudrateList.setOnItemSelectedListener(baudrateSelectedListener);
            baudrateList.setSelection(2);

            // Create an ArrayAdapter using the string array and a default spinner layout
            ArrayAdapter<CharSequence> regionadapter = ArrayAdapter.createFromResource(this, R.array.regions_array, android.R.layout.simple_spinner_item);
            // Specify the layout to use when the list of choices appears
            regionadapter.setDropDownViewResource(android.R.layout.simple_spinner_dropdown_item);
            // Apply the adapter to the spinner
            regionList.setAdapter(regionadapter);
            regionList.setSelection(15);

            tagListAdapter = new TagListAdapter(getApplicationContext(), this);

            tagListView.setLayoutManager(new LinearLayoutManager(this));
            tagListView.addItemDecoration(new DividerItemDecoration(this, DividerItemDecoration.VERTICAL));
            tagListView.setAdapter(tagListAdapter);



            firmwareListener = new FirmwareListener(this);

            loadReadOptionFromProperties();

            setPortraitTableLayout();
            System.gc();
            Thread.currentThread().setDefaultUncaughtExceptionHandler(new GlobalExceptionHandler(activity));
//            RFIDModulePowerCtrl(false);

        } catch (Exception ex) {
            ex.printStackTrace();
        }

        createBeeper();
    }

    @Override
    protected void onResume() {
        registerScanServiceBroadcastReceiver();
        enableLocalReceiver(true);

        if (remapThread.getStatus() == AsyncTask.Status.RUNNING) {
            Log.d(TAG, "remapThread is running");

            if (onPauseTrigger) {
                onPauseTrigger = false;
                targetGunForRFID = !enableBarcode.isChecked();
                RFIDModulePowerCtrl(true);
                Log.d(TAG, "Change target gun to " + (targetGunForRFID ? "RFID" : "Scanner"));
            }
        } else {
            Log.d(TAG, "remapThread is not running");
            onPauseTrigger = false;

            if (skipCount > 0) {
                Log.d(TAG, "skipCount is not 0");
                skipCount--;
            } else {
                if (enableBarcode != null) {
                    //Change the key to internal trigger key to control barcode/RFID
                    targetGunForRFID = !enableBarcode.isChecked();
                    setGunForRFID();
                }
                RFIDModulePowerCtrl(true);
            }
        }

        super.onResume();
    }

    @Override
    protected void onPause() {
        if (firmwareListener != null && firmwareListener.isFirmwareUpgrading()) {
            super.onPause();
            return;
        }

        unregisterReceiver(scanServiceReceiver);
        enableLocalReceiver(false);

        if (getTopActivityName().equalsIgnoreCase("com.android.launcher3.Launcher")) {
            if (serviceListener != null && serviceListener.isReading()) {
                serviceListener.stopReading();
                serviceListener.clearTagRecords();
            }
            onPauseTrigger = true;
            targetGunForRFID = false;
            setGunForRFID();
            if (connectButton != null) {
                if (getConnectBtnText().equals(getString(R.string.Disconnect))) {
                    connectButton.performClick();
                }
            }

            RFIDModulePowerCtrl(false);
        } else if (getTopActivityName().equalsIgnoreCase("com.unitech.rfidreader.ReaderActivity")) {
            if (remapThread.getStatus() == AsyncTask.Status.RUNNING) {
            } else {
                if (skipCount == 0) {
                    if (serviceListener != null && serviceListener.isReading()) {
                        serviceListener.stopReading();
                        serviceListener.clearTagRecords();
                    }
                    onPauseTrigger = true;
                    targetGunForRFID = false;
                    setGunForRFID();
                    if (connectButton != null) {
                        if (getConnectBtnText().equals(getString(R.string.Disconnect))) {
                            connectButton.performClick();
                        }
                    }

                    RFIDModulePowerCtrl(false);
                }
            }
        }
        super.onPause();
    }

    @Override
    protected void onDestroy() {
        super.onDestroy();
    }

    @Override
    protected void onActivityResult(int requestCode, int resultCode, Intent data) {
        switch (requestCode) {
            case FirmwareListener.FILE_SELECT_CODE:
                if (resultCode == RESULT_OK) {
                    firmwareListener.onFileSelectorResult(data);
                }
                break;
        }
        super.onActivityResult(requestCode, resultCode, data);
    }

    //==============================================================================================
    @SuppressLint("InflateParams")
    private void loadSettingContentView() {
        ExpandableListView expandableList = (ExpandableListView) findViewById(R.id.expandableListView1);

        LayoutInflater infalInflater = (LayoutInflater) getSystemService(Context.LAYOUT_INFLATER_SERVICE);
        assert infalInflater != null;
        readOptions = infalInflater.inflate(R.layout.settings_read_options, expandableList, false);
        firmware = infalInflater.inflate(R.layout.firmware, expandableList, false);
        about = infalInflater.inflate(R.layout.about, expandableList, false);
        ArrayList<View> ChildViews = new ArrayList<>();
        ChildViews.add(readOptions);
        ChildViews.add(firmware);
        ChildViews.add(about);

        String[] groups = {"Read Options", "Firmware", "About"};
        expandableListAdapter = new ExpandableListAdapter(this, groups, ChildViews);

        // Set this blank adapter to the list view
        expandableList.setAdapter(expandableListAdapter);
        homeButton.setOnClickListener(new OnClickListener() {
            @Override
            public void onClick(View view) {
//                mContentLoaded = !mContentLoaded;
                showSettingsOrTagsView(0);

                saveReadOptionToProperties();

                if (connectButton != null
                        && getConnectBtnText().equals(getString(R.string.Disconnect))
                        && getReadBtnText().equals(getString(R.string.start_reading))
                        && !firmwareListener.isFirmwareUpgrading()) {
                    setupReader();
                }
            }
        });
        homeButton1.setOnClickListener(new OnClickListener() {
            @Override
            public void onClick(View view) {
                if (serviceListener != null && serviceListener.isRunning()) {
                    return;
                }
                showSettingsOrTagsView(0);
                if (connectButton != null
                        && getConnectBtnText().equals(getString(R.string.Disconnect))
                        && getReadBtnText().equals(getString(R.string.start_reading))
                        && !firmwareListener.isFirmwareUpgrading()) {

                    serviceListener.clearTagRecords();

                    setupReader();
                }
                sendBroadcast(new Intent("ACTION_INITIAL"));

            }
        });
    }

    public void updateSettings() {
        updateRFIDVersion();
        setupReader();
    }

    /**
     * Save Read Options from UI to properties
     */
    public void saveReadOptionToProperties() {
        LoggerUtil.debug(TAG, "saveReadOptionToProperties");

        properties.setOperationMode(operationModeList.getSelectedItemPosition());

        if (operationModeList.getSelectedItemPosition() != Custom_Mode) {
            return;
        }

        properties.setReaderTimeout(Integer.parseInt(readTimeoutView.getText().toString()));

        properties.setAntennaOnTime(Integer.parseInt(antennaOnTimeView.getText().toString()));
        properties.setAntennaOffTime(Integer.parseInt(antennaOffTimeView.getText().toString()));

        properties.setPower(powerList.getSelectedItemPosition());

        properties.setBLF(blfList.getSelectedItemPosition());

        properties.setTARI(tariList.getSelectedItemPosition());

        properties.setSession(sessionList.getSelectedItemPosition());

        properties.setEncoding(encodingList.getSelectedItemPosition());

        properties.setTarget(targetList.getSelectedItemPosition());

        properties.setQValue(qList.getSelectedItemPosition());

        properties.setFastSearch(switchFastSearch.isChecked());
    }

    /**
     * Change Read Option to current properties setting
     */
    private void loadReadOptionFromProperties() {
        LoggerUtil.debug(TAG, "loadReadOptionFromProperties");
        operationModeList.setSelection(properties.getOperationMode());

        loadReadOptionFromProperties(properties.getOperationMode());
    }

    /**
     * Change the Read Option with specific mode from properties
     *
     * @param mode The Operation Mode
     */
    private void loadReadOptionFromProperties(int mode) {
        LoggerUtil.debug(TAG, "loadReadOptionFromProperties with mode " + mode);
        readTimeoutView.setText(String.valueOf(properties.getReaderTimeout(mode)));

        antennaOnTimeView.setText(String.valueOf(properties.getAntennaOnTime(mode)));
        antennaOffTimeView.setText(String.valueOf(properties.getAntennaOffTime(mode)));

        powerList.setSelection(properties.getPower(mode));

        blfList.setSelection(properties.getBLF(mode));

        tariList.setSelection(properties.getTARI(mode));

        sessionList.setSelection(properties.getSession(mode));

        encodingList.setSelection(properties.getEncoding(mode));

        targetList.setSelection(properties.getTarget(mode));

        qList.setSelection(properties.getQValue(mode));

        switchFastSearch.setChecked(properties.getFastSearch(mode));

        lockReaderSetting(mode == Custom_Mode);
    }

    public void lockReaderSetting(boolean enable) {
        readTimeoutView.setEnabled(enable);
        antennaOnTimeView.setEnabled(enable);
        antennaOffTimeView.setEnabled(enable);
        powerList.setEnabled(enable);
        blfList.setEnabled(enable);

        if (operationModeList.getSelectedItemPosition() == Custom_Mode && blfList.getSelectedItemPosition() != BLF_640K)
            tariList.setEnabled(true);
        else
            tariList.setEnabled(false);

        sessionList.setEnabled(enable);
        encodingList.setEnabled(enable);
        targetList.setEnabled(enable);
        qList.setEnabled(enable);
        switchFastSearch.setEnabled(enable);

        if (enable) {
            if (Integer.parseInt(antennaOffTimeView.getText().toString()) < AntennaOffTimeMin) {
                Log.d(TAG, "Change OffTime to " + R.string.antennaOffTimeMin);
                antennaOffTimeView.setText(R.string.antennaOffTimeMin);
            }
        }
    }

    public void closeSettings() {
        showSettingsOrTagsView(0);
    }

    /**
     * Show/Hide the content
     *
     * @param contentLoaded 0 for Main, 1 for Setting, 2 for Action
     */
    public void showSettingsOrTagsView(int contentLoaded) {
        // Decide which view to hide and which to show.
        switch (contentLoaded) {
            case 0:
                mDisplayView.setVisibility(View.VISIBLE);
                mSettingView.setVisibility(View.GONE);
                mActionView.setVisibility(View.GONE);
                showSaveTagBtn(true);
                break;
            case 1:
                mDisplayView.setVisibility(View.GONE);
                mSettingView.setVisibility(View.VISIBLE);
                mActionView.setVisibility(View.GONE);
                showSaveTagBtn(false);
                break;
            case 2:
                mDisplayView.setVisibility(View.GONE);
                mSettingView.setVisibility(View.GONE);
                mActionView.setVisibility(View.VISIBLE);
                showSaveTagBtn(false);
                break;
        }
    }


    /**
     * This is called when the screen rotates.
     * (onCreate is no longer called when screen rotates due to manifest, see: android:configChanges)
     */
    @Override
    public void onConfigurationChanged(@NonNull Configuration newConfig) {
        super.onConfigurationChanged(newConfig);
        int orientation = newConfig.orientation;
        if (orientation == Configuration.ORIENTATION_PORTRAIT) {
            setPortraitTableLayout();
        } else if (orientation == Configuration.ORIENTATION_LANDSCAPE) {
            setPortraitTableLayout();
        }
    }

    private void setPortraitTableLayout() {
        tableNumberWidth = (int) (windowWidth * (0.12));
        tableEPCWidth = (int) (windowWidth * (0.6));
        tableRSSIWidth = (int) (windowWidth * (0.12));
        tableCountWidth = (int) (windowWidth * (0.16));

        tableNumberView.setWidth(tableNumberWidth);
        tableEPCView.setWidth(tableEPCWidth);
        tableRSSIView.setWidth(tableRSSIWidth);
        tableCountView.setWidth(tableCountWidth);
    }

    @Override
    public void onBackPressed() {
    }

    @Override
    public boolean onKeyDown(int keyCode, KeyEvent event) {
        if (enableBarcode.isChecked()) {
            return super.onKeyDown(keyCode, event);
        }

        if (keyCode == KEYCODE_SCAN_EXTENDED) {
            if (ReaderActivity.this.reader != null && getConnectBtnText().equals(getString(R.string.Disconnect))
                    && serviceListener != null && !serviceListener.isRunning()) {
                //Harvey should setup the Inventory parameter for locate
                if (mActionView.getVisibility() == View.VISIBLE) {
                    if (ActionTabPosition != 4) {
                        return true;
                    }
                    String targetTag = sectionsStateAdapter.getEPCString("locate");

                    if (targetTag.trim().equalsIgnoreCase("")) {
                        Toast.makeText(this, "Please input target tag!", Toast.LENGTH_SHORT).show();
                        return true;

                    } else {
                        try {
                            setTargetTag(targetTag);
                        } catch (ReaderException e) {
                            e.printStackTrace();
                        }
                    }
                }
                serviceListener.startReading();
            }
            return true;
        } else {
            return super.onKeyDown(keyCode, event);
        }
    }

    @Override
    public boolean onKeyLongPress(int keyCode, KeyEvent event) {
        if (enableBarcode.isChecked()) {
            return super.onKeyLongPress(keyCode, event);
        }

        if (keyCode == KEYCODE_SCAN_EXTENDED) {
            if (ReaderActivity.this.reader != null && getConnectBtnText().equals(getString(R.string.Disconnect))
                    && serviceListener != null && !serviceListener.isReading()) {
                //Harvey should setup the Inventory parameter for locate
//                    if (mActionView.getVisibility() == View.VISIBLE) {
//                        setupReaderForLocate();
//                    }
//                    serviceListener.startReading();
            }
            return true;
        } else {
            return super.onKeyLongPress(keyCode, event);
        }
    }

    @Override
    public boolean onKeyUp(int keyCode, KeyEvent event) {
        if (enableBarcode.isChecked()) {
            return super.onKeyDown(keyCode, event);
        }
        if (keyCode == KEYCODE_SCAN_EXTENDED) {
            //                if (serviceListener != null && serviceListener.isReading()) {
            serviceListener.stopReading();
//                }
            return true;
        } else {
            return super.onKeyUp(keyCode, event);
        }
    }

    private void findAllViewsById() {
        getWindowManager().getDefaultDisplay().getMetrics(displaymetrics);

        //region Button
        connectButton = findViewById(R.id.btn_connect);
        scanButton = findViewById(R.id.btn_scan);
        clearButton = findViewById(R.id.btn_Clear);
        settingsButton = findViewById(R.id.btn_setting);
        homeButton = findViewById(R.id.btn_back_main);
        homeButton1 = findViewById(R.id.btn_back_main2);
        tagOperationButton = findViewById(R.id.btn_tag_operation);
        savetagButton = findViewById(R.id.btn_savetag);
        //endregion

        //region TextView
        searchResultCount = findViewById(R.id.search_result_view);
        tableNumberView = findViewById(R.id.title_no);
        tableEPCView = findViewById(R.id.title_epc);
        tableRSSIView = findViewById(R.id.title_rssi);
        tableCountView = findViewById(R.id.title_count);
        //endregion

        read_result_layout = findViewById(R.id.read_result_layout);

        regionList = findViewById(R.id.Region_List);
        baudrateList = findViewById(R.id.spinner_baudrate);

        checkBeep = findViewById(R.id.BeepControl);
        checkBeep.setOnCheckedChangeListener(beeperCheckedListener);

        enableBarcode = findViewById(R.id.BarcodeControl);
        enableBarcode.setOnCheckedChangeListener(scanServiceCheckedListener);
        barcodeText = findViewById(R.id.barcodeText);
        barcodeText.setRawInputType(InputType.TYPE_NULL);

//        tableTags = findViewById(R.id.table_tags);


        temperatureView = findViewById(R.id.temperature);
        avgReadView = findViewById(R.id.avg_reads_view);
        totalTagCountView = findViewById(R.id.totalTagCount_view);
        readingTimeView = findViewById(R.id.reading_time);
        errorMsgView = findViewById(R.id.error_messages_view);
//        maxReadRateView = findViewById(R.id.max_read_rate);

//        tagTable = findViewById(R.id.tablelayout);
        tagListView = findViewById(R.id.recycler_view);

        //ActionView
        tabLayout = (TabLayout) findViewById(R.id.tabs);
        viewPager = (ViewPager2) findViewById(R.id.view_pager);


        //-----------------------------------------------------------------------------------------
    }

    private class PositionOnTabSelectedListener implements TabLayout.OnTabSelectedListener {
        @Override
        public void onTabSelected(@NonNull TabLayout.Tab tab) {
            ActionTabPosition = tab.getPosition();
            LoggerUtil.debug(TAG, "ActionTabPosition = " + ActionTabPosition);
//            if (ActionTabPosition == 4)
            setupReaderForLocate();
//            else if (ActionTabPosition == 0)
//            {
//                setupReaderForLocate();
//                setupReader(2000, 600);
//            }
        }

        @Override
        public void onTabUnselected(TabLayout.Tab tab) {
            // No-op
        }

        @Override
        public void onTabReselected(TabLayout.Tab tab) {
            // No-op
        }
    }

    private void findReadOptionAllViewsById() {
        powerList = this.readOptions.findViewById(R.id.antennaPower);

        operationModeList = readOptions.findViewById(R.id.spinner_operation_mode);

        readTimeoutView = readOptions.findViewById(R.id.read_time_value);
        antennaOnTimeView = readOptions.findViewById(R.id.edit_async_on_time);
        antennaOffTimeView = readOptions.findViewById(R.id.edit_async_off_time);

        blfList = readOptions.findViewById(R.id.spinner_blf);
        tariList = readOptions.findViewById(R.id.spinner_tari);
        sessionList = readOptions.findViewById(R.id.spinner_session);
        encodingList = readOptions.findViewById(R.id.spinner_tag_encoding);
        targetList = readOptions.findViewById(R.id.spinner_target);
        qList = readOptions.findViewById(R.id.spinner_q);
        switchFastSearch = readOptions.findViewById(R.id.switch_fast_search);

        initUIReadOption();
    }

    private void findFirmwareAllViewsById() {
        layout_firmware = firmware.findViewById(R.id.layout_firmware);
        firmware_hint = firmware.findViewById(R.id.firmware_hint);
        softwareVersionView = firmware.findViewById(R.id.version_software);
        hardwareVersionView = firmware.findViewById(R.id.version_hardware);
        modelVersionView = firmware.findViewById(R.id.version_model);
        productGroupVersionView = firmware.findViewById(R.id.version_product_group);
        serialNumberView = firmware.findViewById(R.id.version_serial);
        firmware_hint = firmware.findViewById(R.id.firmware_hint);
    }


    private void initTagOperation() {
        initTagOperationFlag = true;
        final String[] tab_text_array = getResources().getStringArray(R.array.tab_text_array);
        tabLayout.setTabTextColors(Color.WHITE, Color.WHITE);
        tabLayout.setSelectedTabIndicatorColor(Color.rgb(255, 128, 128));
        tabLayout.addOnTabSelectedListener(new PositionOnTabSelectedListener());
        TabLayoutMediator.TabConfigurationStrategy tabConfigurationStrategy = new TabLayoutMediator.TabConfigurationStrategy() {
            @Override
            public void onConfigureTab(@NonNull TabLayout.Tab tab, int position) {
                tab.setText(tab_text_array[position]);
            }
        };
        sectionsStateAdapter = new SectionsStateAdapter(getSupportFragmentManager(), getLifecycle());
        for (String s : tab_text_array) {
            sectionsStateAdapter.addFragment(new PlaceholderFragment(this, s));
        }
        viewPager.setOffscreenPageLimit(4);
        viewPager.setAdapter(sectionsStateAdapter);
        new TabLayoutMediator(tabLayout, viewPager, tabConfigurationStrategy).attach();
    }

    private void initUIReadOption() {
        // Create an ArrayAdapter using the string array and a default spinner layout
        ArrayAdapter<CharSequence> poweradapter = ArrayAdapter.createFromResource(this, R.array.powerlevel_array, android.R.layout.simple_spinner_item);
        // Specify the layout to use when the list of choices appears
        poweradapter.setDropDownViewResource(android.R.layout.simple_spinner_dropdown_item);
        // Apply the adapter to the spinner
        powerList.setAdapter(poweradapter);
        powerList.setSelection(20);

        ArrayAdapter<CharSequence> operationMode_Adapter = ArrayAdapter.createFromResource(this, R.array.operation_mode_array, android.R.layout.simple_spinner_item);
        operationMode_Adapter.setDropDownViewResource(android.R.layout.simple_spinner_dropdown_item);
        operationModeList.setAdapter(operationMode_Adapter);
        operationModeList.setOnItemSelectedListener(operationModeSelectedListener);
        operationModeList.setSelection(defaultMode);

        antennaOnTimeView.setOnFocusChangeListener(asyncOnTimeFocusChangeListener);
        antennaOffTimeView.setOnFocusChangeListener(asyncOffTimeFocusChangeListener);

        ArrayAdapter<CharSequence> blfAdapter = ArrayAdapter.createFromResource(this, R.array.backscatter_link_frequency_array, android.R.layout.simple_spinner_item);
        blfAdapter.setDropDownViewResource(android.R.layout.simple_spinner_dropdown_item);
        blfList.setAdapter(blfAdapter);
        blfList.setOnItemSelectedListener(blfSelectedListener);

        ArrayAdapter<CharSequence> tariAdapter = ArrayAdapter.createFromResource(this, R.array.tari_array, android.R.layout.simple_spinner_item);
        tariAdapter.setDropDownViewResource(android.R.layout.simple_spinner_dropdown_item);
        tariList.setAdapter(tariAdapter);

        ArrayAdapter<CharSequence> sessionAdapter = ArrayAdapter.createFromResource(this, R.array.session_array, android.R.layout.simple_spinner_item);
        sessionAdapter.setDropDownViewResource(android.R.layout.simple_spinner_dropdown_item);
        sessionList.setAdapter(sessionAdapter);

        ArrayAdapter<CharSequence> tagEncodingAdapter = ArrayAdapter.createFromResource(this, R.array.tag_encoding_array, android.R.layout.simple_spinner_item);
        tagEncodingAdapter.setDropDownViewResource(android.R.layout.simple_spinner_dropdown_item);
        encodingList.setAdapter(tagEncodingAdapter);

        ArrayAdapter<CharSequence> targetAdapter = ArrayAdapter.createFromResource(this, R.array.target_array, android.R.layout.simple_spinner_item);
        targetAdapter.setDropDownViewResource(android.R.layout.simple_spinner_dropdown_item);
        targetList.setAdapter(targetAdapter);

        ArrayAdapter<CharSequence> qAdapter = ArrayAdapter.createFromResource(this, R.array.target_q, android.R.layout.simple_spinner_item);
        qAdapter.setDropDownViewResource(android.R.layout.simple_spinner_dropdown_item);
        qList.setAdapter(qAdapter);
    }

    //region UI Listener
    private View.OnFocusChangeListener asyncOnTimeFocusChangeListener = new View.OnFocusChangeListener() {
        @Override
        public void onFocusChange(View v, boolean hasFocus) {
            if (!hasFocus) {
                if (Integer.parseInt(antennaOnTimeView.getText().toString()) > AntennaOnTimeMax) {
                    Log.d(TAG, "Change OnTime to " + R.string.antennaOnTimeMax);
                    antennaOnTimeView.setText(R.string.antennaOnTimeMax);
                }
            }
        }
    };

    private View.OnFocusChangeListener asyncOffTimeFocusChangeListener = new View.OnFocusChangeListener() {
        @Override
        public void onFocusChange(View v, boolean hasFocus) {
            if (!hasFocus) {
                if (Integer.parseInt(antennaOffTimeView.getText().toString()) < AntennaOffTimeMin) {
                    Log.d(TAG, "Change OffTime to " + R.string.antennaOffTimeMin);
                    antennaOffTimeView.setText(R.string.antennaOffTimeMin);
                }
            }
        }
    };

    private AdapterView.OnItemSelectedListener baudrateSelectedListener = new AdapterView.OnItemSelectedListener() {
        @Override
        public void onItemSelected(AdapterView<?> parent, View view, int position, long id) {
            switch (position) {
                case 0:
                    ReaderConnect.setBaudrate(115200);
                    break;
                case 1:
                    ReaderConnect.setBaudrate(230400);
                    break;
                default:
                case 2:
                    ReaderConnect.setBaudrate(460800);
                    break;
                case 3:
                    ReaderConnect.setBaudrate(921600);
                    break;
            }
        }

        @Override
        public void onNothingSelected(AdapterView<?> parent) {

        }
    };

    private AdapterView.OnItemSelectedListener operationModeSelectedListener = new AdapterView.OnItemSelectedListener() {

        @Override
        public void onItemSelected(AdapterView<?> parent, View view, int position, long id) {
            loadReadOptionFromProperties(position);
        }

        @Override
        public void onNothingSelected(AdapterView<?> parent) {

        }
    };

    private AdapterView.OnItemSelectedListener blfSelectedListener = new AdapterView.OnItemSelectedListener() {
        @Override
        public void onItemSelected(AdapterView<?> parent, View view, int position, long id) {
            switch (position) {
                case BLF_250K:
                    if (operationModeList.getSelectedItemPosition() == Custom_Mode) {
                        tariList.setEnabled(true);
                    }
                    break;
                case BLF_640K:
                    if (operationModeList.getSelectedItemPosition() == Custom_Mode) {
                        tariList.setSelection(TARI_6_25);
                        tariList.setEnabled(false);
                    }
                    break;
            }
        }

        @Override
        public void onNothingSelected(AdapterView<?> parent) {

        }
    };

    private CheckBox.OnCheckedChangeListener beeperCheckedListener = new CheckBox.OnCheckedChangeListener() {
        @Override
        public void onCheckedChanged(CompoundButton buttonView, boolean isChecked) {
            if (isChecked) {
                mBeeper.setCanBeep(true);
                mBeeper.stopAllBeep();
                mBeeper.tryBeep();
            } else {
                mBeeper.stopAllBeep();
                mBeeper.setCanBeep(false);
            }
        }
    };

    private CheckBox.OnCheckedChangeListener scanServiceCheckedListener = new CheckBox.OnCheckedChangeListener() {
        @Override
        public void onCheckedChanged(CompoundButton buttonView, boolean isChecked) {
            setGunKeyViaBarcodeCheckBox(!isChecked);

            if (isChecked) {
                barcodeText.setEnabled(true);
                barcodeText.setVisibility(View.VISIBLE);
                barcodeText.requestFocus();
                barcodeText.setShowSoftInputOnFocus(false);

                if (connectButton != null) {
                    if (getConnectBtnText().equals(getString(R.string.Disconnect))) {
                        connectButton.performClick();
                        preRFIDStatus = true;
                    } else
                        preRFIDStatus = false;
                }

                connectButton.setEnabled(false);
                tagOperationButton.setEnabled(false);

                if (getReadBtnText().equals(getString(R.string.start_reading)) && scanButton.isEnabled()) {
                    scanButton.setEnabled(false);
                    scanButton.setTag(Boolean.TRUE);
                }

                saveTagBtnEdit(false);

                tagListView.setVisibility(View.GONE);
//                tableTags.setVisibility(View.GONE);

                if (read_result_layout.getVisibility() == View.VISIBLE) {
                    read_result_layout.setVisibility(View.GONE);
                    read_result_layout.setTag(Boolean.TRUE);
                }

            } else {
                barcodeText.setEnabled(false);
                barcodeText.setVisibility(View.GONE);
                connectButton.setEnabled(true);
                connectButton.setClickable(true);
                tagOperationButton.setEnabled(true);


                if (connectButton != null) {
                    if (preRFIDStatus) {
                        connectButton.performClick();
                    }
                }
//
//                if (getConnectBtnText().equals(getString(R.string.Connect))) {
//                    scanButton.setEnabled(false);
//                }
//                if (scanButton.getTag() == Boolean.TRUE) {
//                    scanButton.setTag(Boolean.FALSE);
//                    scanButton.setEnabled(true);
//                }
                tagListView.setVisibility(View.VISIBLE);
//                tableTags.setVisibility(View.VISIBLE);

                if (read_result_layout.getTag() == Boolean.TRUE) {
                    read_result_layout.setTag(Boolean.FALSE);
                    read_result_layout.setVisibility(View.VISIBLE);
                }
            }
        }
    };

    public OnClickListener clearListener = new OnClickListener() {
        @Override
        public void onClick(View v) {
            if (enableBarcode.isChecked()) {
                clearBarcodeText();
            } else {
                serviceListener.clearTagRecords();
            }
        }
    };

    //endregion

    public boolean createDirectory(String pathName) {
        File file = new File(pathName);

        if (file.exists())
            return true;
        else {
            if (file.mkdirs())
                return true;
            else {
                if (createDirectory(file.getParent()))
                    return file.mkdirs();
                else
                    return false;
            }
        }
    }

    //region UI Control Functions

    /**
     * Reset the UI to clear
     */
    public void initUI() {
        setSearchResultCountText("");
        setTemperatureText("");
        setAvgReadText("");
        setTotalTagCountText("");
        setReadingTimeText("");
//        setMaxReadRate("");
//        setErrorMsg("");

        resetTagTable();
    }

    public void connectBtnEdit(boolean enable) {
        connectButton.setEnabled(enable);
        connectButton.setClickable(enable);
    }

    public void readBtnEdit(boolean enable) {
        scanButton.setEnabled(enable);
        scanButton.setClickable(enable);
    }

    public void tagOpBtnEdit(boolean enable) {
        if (enable && !initTagOperationFlag) {
            initTagOperation();
        }
        tagOperationButton.setEnabled(enable);
        tagOperationButton.setClickable(enable);
    }

    public void beepCheckBoxEdit(boolean enable) {
        checkBeep.setEnabled(enable);
        checkBeep.setClickable(enable);
    }

    public void saveTagBtnEdit(boolean enable) {
        savetagButton.setEnabled(enable);
        savetagButton.setClickable(enable);
    }

    public void showSaveTagBtn(boolean show) {
        savetagButton.setVisibility(show ? View.VISIBLE : View.INVISIBLE);
    }

    public void regionListEdit(boolean enable) {
        regionList.setEnabled(enable);
    }

    public String getSelectedRegionList() {
        return regionList.getSelectedItem().toString();
    }

    public void baudrateListEdit(boolean enable) {
        baudrateList.setEnabled(enable);
    }

    public void setConnectBtnText(String text) {
        connectButton.setText(text);
    }

    public String getConnectBtnText() {
        return connectButton.getText().toString();
    }

    public void setReadBtnText(String text) {
        scanButton.setText(text);
    }

    public String getReadBtnText() {
        return scanButton.getText().toString();
    }

    public void setTemperatureText(String temperature) {
        temperatureView.setText(temperature);
    }

    public void setTemperatureText(Spanned fromHtml) {
        temperatureView.setText(fromHtml);
    }

//    public void setMaxReadRate(String maxReadRate){
//        maxReadRateView.setText(maxReadRate);
//    }
//
//    public void setMaxReadRate(Spanned fromHtml) {
//        maxReadRateView.setText(fromHtml);
//    }

    public void setAvgReadText(String avgRead) {
        avgReadView.setText(avgRead);
    }

    public void setAvgReadText(Spanned fromHtml) {
        avgReadView.setText(fromHtml);
    }

    public void setSearchResultCountText(String count) {
        searchResultCount.setText(count);
    }

    public void setSearchResultCountText(Spanned fromHtml) {
        searchResultCount.setText(fromHtml);
    }

    public void setSearchResultTextColor(int color) {
        searchResultCount.setTextColor(color);
    }

    public int getSearchResultTextColor() {
        return searchResultCount.getTextColors().getDefaultColor();
    }

    public void setTotalTagCountText(String count) {
        totalTagCountView.setText(count);
    }

    public void setTotalTagCountText(Spanned fromHtml) {
        totalTagCountView.setText(fromHtml);
    }

    public void setReadingTimeText(String time) {
        readingTimeView.setText(time);
    }

    public void setReadingTimeText(Spanned fromHtml) {
        readingTimeView.setText(fromHtml);
    }

    public void resetTagTable() {
//        tagTable.removeAllViews();
//        ArrayList<TagRecord> tr = new ArrayList<>();
//        TagListAdapter tags = new TagListAdapter(getApplicationContext(), this, tr);
//        tagListView.setAdapter(tags);
        tagListAdapter.clear();
    }

    public void setErrorMsg(String msg) {
        errorMsgView.setText(msg);
    }

    public void setErrorMsgTextColor(int color) {
        errorMsgView.setTextColor(color);
    }

    public void showErrorMsg(boolean show) {
        errorMsgView.setVisibility(show ? View.VISIBLE : View.GONE);
    }

    public void showLayoutFirmware(boolean show) {
        layout_firmware.setVisibility(show ? View.VISIBLE : View.GONE);
    }

    public void showFirmwareHint(boolean show) {
        firmware_hint.setVisibility(show ? View.VISIBLE : View.GONE);
    }

    public void UpdateTagList(ConcurrentHashMap<String, TagRecord> tagList) {
        //        tagListAdapter.UpdateData(data);
//        TagListAdapter tagListAdapter = new TagListAdapter(getApplicationContext(), this, data);
//        tagListView.setAdapter(tagListAdapter);
        tagListAdapter.add(tagList);
    }

    //endregion

    //region RFID Power Control
    public void RFIDModulePowerCtrl(final boolean bEnable) {

        SetRFIDPowerThread setRFIDPowerThread = new SetRFIDPowerThread(this, bEnable);
        setRFIDPowerThread.executeOnExecutor(AsyncTask.THREAD_POOL_EXECUTOR);
    }

    private class SetRFIDPowerThread extends AsyncTask<Void, Void, String> {
        ReaderActivity _activity;
        boolean _enable = false;

        SetRFIDPowerThread(ReaderActivity activity, boolean enable) {
            _activity = activity;
            _enable = enable;
        }

        @Override
        protected void onPreExecute() {
            super.onPreExecute();
        }

        @Override
        protected String doInBackground(Void... voids) {
            try {
                PogoCtrl potoCtrl = PogoCtrl.getInstance(_activity.getApplicationContext());

                if (_enable)
                    potoCtrl.powerOn();
                else
                    potoCtrl.powerOff();

            } catch (final Exception ex) {
                LoggerUtil.error("RFID Power Control", "Get exception : " + ex.getMessage());
                _activity.runOnUiThread(new Runnable() {
                    @Override
                    public void run() {
                        _activity.showErrorMsg(true);
                        _activity.setErrorMsgTextColor(0xffff0000);
                        _activity.setErrorMsg(ex.getMessage());
                        connectBtnEdit(false);
                    }
                });
            }
            return null;
        }

        @Override
        protected void onPostExecute(String s) {
            super.onPostExecute(s);
        }
    }
    //endregion

    private void createBeeper() {
        mBeeper = Beeper.getInstance();
        if (checkBeep != null) {
            mBeeper.setCanBeep(checkBeep.isChecked());
        }
    }

    public void clearBarcodeText() {
        this.barcodeText.getText().clear();
    }

    public void disableBarcode() {
        this.enableBarcode.setEnabled(false);
    }

    public void enableBarcode() {
        this.enableBarcode.setEnabled(true);
    }

    private void registerScanServiceBroadcastReceiver() {
        IntentFilter scannerIntentFilter = new IntentFilter();
        scannerIntentFilter.addAction("unitech.scanservice.data");
        scannerIntentFilter.addAction("unitech.scanservice.databytelength");
        registerReceiver(scanServiceReceiver, scannerIntentFilter);
    }

    private void enableLocalReceiver(boolean enable) {
        if (enable) {
            IntentFilter filter = new IntentFilter();
            filter.addAction(Action_SelectEPC);
            LocalBroadcastManager.getInstance(this).registerReceiver(mLocalBroadcastReceiver, filter);
//            registerReceiver(mLocalBroadcastReceiver, filter);
        } else {
            LocalBroadcastManager.getInstance(this).unregisterReceiver(mLocalBroadcastReceiver);
        }
    }

    /**
     * Set the key code of SCAN_GUN
     */
    private void setGunForRFID() {
        if (remapThread.getStatus() != AsyncTask.Status.RUNNING) {
            if (remapThread.getStatus() == AsyncTask.Status.FINISHED) {
                remapThread = new SetGunKeyViaResumePauseThread("Switching gun key...", this);
            } else {
                remapThread.setMessage("Switching gun key...");
            }
            remapThread.executeOnExecutor(AsyncTask.THREAD_POOL_EXECUTOR);
//            remapThread.execute();
        }
    }

    private void setGunKeyViaBarcodeCheckBox(boolean isExtended) {
        SetGunKeyViaBarcodeCheckBoxThread newRemapThread = new SetGunKeyViaBarcodeCheckBoxThread("Switching gun key...", this, isExtended);
        newRemapThread.executeOnExecutor(AsyncTask.THREAD_POOL_EXECUTOR);
//        newRemapThread.execute();
    }

    private class SetGunKeyViaResumePauseThread extends AsyncTask<Void, Void, String> {
        private ReaderActivity mReaderActivity;
        private String _message;

        SetGunKeyViaResumePauseThread(ReaderActivity activity) {
            this.mReaderActivity = activity;
        }

        SetGunKeyViaResumePauseThread(String message, ReaderActivity activity) {
            this.mReaderActivity = activity;
            this._message = message;
        }

        @Override
        protected void onPreExecute() {
            keyRemapDialog.setMessage(_message);
            keyRemapDialog.show();
        }

        @Override
        protected String doInBackground(Void... params) {
            String exception = "Exception :";
            boolean keyType = false;

            do {
                if (mReaderActivity.firstExecute) {
                    if (!exportProgrammableKey()) {
                        LoggerUtil.error(TAG, "Export programmable key fail\r\n");
                        exception += "Export programmable key fail\r\n";
                        return exception;
                    }
                } else {
                    if (!isKeyConfigFileExist()) {
                        LoggerUtil.debug(TAG, "Does not find key config file, going to export...");

                        if (!exportProgrammableKey()) {
                            LoggerUtil.error(TAG, "Export programmable key fail\r\n");
                            exception += "Export programmable key fail\r\n";
                            return exception;
                        }
                    }
                }

                keyType = targetGunForRFID;
                boolean enableKeyRemap = false;

                if (!keyType && onPauseTrigger) {
                    Log.v(TAG, "Will set Key Remap to default value : " + defaultKeyRemapEnable);
                    enableKeyRemap = defaultKeyRemapEnable;
                } else {
                    Log.v(TAG, "Will enable Key Remap");
                    enableKeyRemap = true;
                }

                if (!setKeyRemapStatus(enableKeyRemap)) {
                    exception += "Set key remap status fail\r\n";
                    return exception;
                }

                if (!setTriggerKeyCode(keyType)) {
                    exception += "Set trigger key code fail\r\n";
                    return exception;
                }
            } while (keyType != targetGunForRFID);

            return exception;
        }

        @Override
        protected void onPostExecute(String exception) {
            keyRemapDialog.dismiss();
        }

        public void setMessage(String message) {
            _message = message;
        }
    }

    private class SetGunKeyViaBarcodeCheckBoxThread extends AsyncTask<Void, Void, String> {
        private ReaderActivity mReaderActivity;
        private String _message = "";
        private boolean isExtended = false;

        SetGunKeyViaBarcodeCheckBoxThread(String message, ReaderActivity activity, boolean extended) {
            this.mReaderActivity = activity;
            this._message = message;
            isExtended = extended;
        }

        @Override
        protected void onPreExecute() {
            keyRemapDialog.setMessage(_message);
            keyRemapDialog.show();
        }

        @Override
        protected String doInBackground(Void... params) {
            String exception = "Exception :";

            if (!isKeyConfigFileExist()) {
                LoggerUtil.debug(TAG, "Does not find key config file, going to export...");

                if (!exportProgrammableKey()) {
                    LoggerUtil.error(TAG, "Export programmable key fail\r\n");
                    exception += "Export programmable key fail\r\n";
                    return exception;
                }
            }

            if (!setKeyRemapStatus(true)) {
                exception += "Set key remap status fail\r\n";
                return exception;
            }

            if (!setTriggerKeyCode(isExtended)) {
                exception += "Set trigger key code fail\r\n";
            }

            return exception;
        }

        @Override
        protected void onPostExecute(String exception) {
            keyRemapDialog.dismiss();
        }
    }

    private boolean exportProgrammableKey() {
        try {
            keySetting = KeySetting.getInstance(getApplicationContext());

            if (firstExecute) {
                Bundle result = keySetting.getKeyRemapStatus();
                defaultKeyRemapEnable = result.getBoolean(KeySetting.BUNDLE_KEY_REMAPPING);
                result = keySetting.getRFIDTriggerKeyCode();
                KEYCODE_SCAN_EXTENDED = Integer.valueOf(result.getString(KeySetting.BUNDLE_KEY_CODE));
                firstExecute = false;
                LoggerUtil.debug(TAG, "Get Default Key Remap Setting : " + defaultKeyRemapEnable);
            }
            return true;
        } catch (Exception e) {
            e.printStackTrace();
            return false;
        }
    }

    private boolean setKeyRemapStatus(boolean enableSetting) {
        try {
            if (getTopActivityName().equalsIgnoreCase("com.unitech.rfidreader.ReaderActivity")) {
                if (!onPauseTrigger) {
                    Log.d(TAG, "onPauseTrigger is false, skipCount = 1");
                    skipCount = 1;
                }
            }

            LoggerUtil.debug(TAG, "Set key remap : " + enableSetting + "\r\n");

            Bundle result = keySetting.setKeyRemapStatus(enableSetting);

            if (result.getInt(KeySetting.BUNDLE_ERROR_CODE) != KeySetting.RESULT_CODE_SUCCESS) {
                LoggerUtil.error(TAG, "Set key remap status fail : " + result.getString(KeySetting.BUNDLE_ERROR_MSG) + "\r\n");
                return false;
            }

            return true;
        } catch (Exception e) {
            e.printStackTrace();
            LoggerUtil.error(TAG, "Set key remap status exception\r\n" + e.getMessage());
            return false;
        }
    }

    private boolean setTriggerKeyCode(boolean isExtended) {
        try {
            if (getTopActivityName().equalsIgnoreCase("com.unitech.rfidreader.ReaderActivity")) {
                if (!onPauseTrigger) {
                    Log.d(TAG, "onPauseTrigger is false, skipCount = 1");
                    skipCount = 1;
                }
            }

            LoggerUtil.debug(TAG, "Set trigger key : " + isExtended + "\r\n");

            Bundle result = keySetting.changeTriggerKeyMode(isExtended);

            if (result.getInt(KeySetting.BUNDLE_ERROR_CODE) != KeySetting.RESULT_CODE_SUCCESS) {
                LoggerUtil.error(TAG, "Set trigger key code fail : " + result.getString(KeySetting.BUNDLE_ERROR_MSG) + "\r\n");
                return false;
            }

            return true;
        } catch (Exception e) {
            e.printStackTrace();
            LoggerUtil.error(TAG, "Set trigger key code exception\r\n" + e.getMessage());
            return false;
        }
    }

    private boolean isKeyConfigFileExist() {
        File fileKeyConfig = new File(FILE_PATH_KEYS_CONFIG);

        return fileKeyConfig.exists();
    }

    private String getTopActivityName() {
        ActivityManager am = (ActivityManager) this.getSystemService(ACTIVITY_SERVICE);
        List<ActivityManager.RunningTaskInfo> taskInfo = am.getRunningTasks(1);

//        Log.d("topActivity", "CURRENT Activity ::" + taskInfo.get(0).topActivity.getClassName());
        ComponentName componentInfo = taskInfo.get(0).topActivity;

//        Log.d("topActivity", "CURRENT PackageName ::" + componentInfo.getPackageName());

        return taskInfo.get(0).topActivity.getClassName();
    }

    public String getVersionName(Context context) {
        PackageManager packageManager = context.getPackageManager();
        PackageInfo packageInfo;
        String versionName = "";
        try {
            packageInfo = packageManager.getPackageInfo(context.getPackageName(), 0);
            versionName = packageInfo.versionName;
        } catch (PackageManager.NameNotFoundException e) {
            e.printStackTrace();
        }
        return versionName;
    }

    public String getTimeFormat(String format) {
        return new SimpleDateFormat(format, Locale.getDefault()).format(new Date());
    }

    private BroadcastReceiver scanServiceReceiver = new BroadcastReceiver() {
        @Override
        public void onReceive(Context context, Intent intent) {
            Bundle bundle;
            if (intent != null) {
                if (intent.getAction() == null)
                    return;

                switch (intent.getAction()) {
                    case "unitech.scanservice.data":
                        //收到條碼資料字串
                        bundle = intent.getExtras();
                        if (bundle != null) {
                            String barcodeData = bundle.getString("text");//取得資料字串
                            Log.d(TAG, String.format("Recv scan str: %s", barcodeData));
                            barcodeText.append(barcodeData);
                        }
                        break;
                    case "unitech.scanservice.databytelength":
                        //收到條碼資料長度
                        bundle = intent.getExtras();
                        if (bundle != null) {
                            int len = bundle.getInt("text");//取得資料長度
                            Log.d(TAG, String.format("Recv scan len: %d", len));
                        }
                        break;
                    default:
                        break;
                }
            }
        }
    };

    private final BroadcastReceiver mLocalBroadcastReceiver = new BroadcastReceiver() {
        @Override
        public void onReceive(Context context, Intent mIntent) {
            String action = mIntent.getAction();
            Bundle bundle = mIntent.getExtras();
            if (action == null) return;
            if (bundle == null) return;
            switch (action) {
                case Action_SelectEPC: {
                    String epc = bundle.getString("epc");

                    if (getReadBtnText().equals(getString(R.string.start_reading)) &&
                            reader != null &&
                            getConnectBtnText().equals(getString(R.string.Disconnect)) &&
                            serviceListener != null
                    ) {
                        targetEPC = epc;
                        sectionsStateAdapter.SetEPCString(epc);
                        sectionsStateAdapter.notifyDataSetChanged();
                        showSettingsOrTagsView(2);
                    }
                    break;
                }
            }
        }
    };


    //region Reader Control

    /**
     * Setup RFID reader with properties
     */
    public void setupReader() {
        int async_on = 150;
        int async_off = 100;
        try {
//            async_on = Integer.parseInt(antennaOnTimeView.getText().toString());
//            async_off = Integer.parseInt(antennaOffTimeView.getText().toString());

            async_on = properties.getAntennaOnTime();
            async_off = properties.getAntennaOffTime();

        } catch (Exception e) {
            e.printStackTrace();
        }
        setupReader(async_on, async_off);
    }

    /**
     * Setup RFID reader with properties
     *
     * @param onTime  Antenna on time
     * @param offTime Antenna off time
     */
    public void setupReader(int onTime, int offTime) {
        try {
            //region Set antenna power
            int powerLevel = (properties.getPower() + 10) * 100;
            setParamToReader("/reader/radio/readPower", powerLevel);
            //endregion

            //region Set read timeout
            setParamToReader("/reader/commandTimeout", properties.getReaderTimeout());
            //endregion

            //region Set BLF
            switch (properties.getBLF()) {
                default:
                case BLF_250K:
                    setParamToReader(TMConstants.TMR_PARAM_GEN2_BLF, Gen2.LinkFrequency.LINK250KHZ);
                    break;
                case BLF_640K:
                    setParamToReader(TMConstants.TMR_PARAM_GEN2_BLF, Gen2.LinkFrequency.LINK640KHZ);
                    break;
            }
            //endregion

            //region Set Session
            switch (properties.getSession()) {
                default:
                case Session_S0:
                    setParamToReader(TMConstants.TMR_PARAM_GEN2_SESSION, Gen2.Session.S0);
                    break;
                case Session_S1:
                    setParamToReader(TMConstants.TMR_PARAM_GEN2_SESSION, Gen2.Session.S1);
                    break;
                case Session_S2:
                    setParamToReader(TMConstants.TMR_PARAM_GEN2_SESSION, Gen2.Session.S2);
                    break;
                case Session_S3:
                    setParamToReader(TMConstants.TMR_PARAM_GEN2_SESSION, Gen2.Session.S3);
                    break;
            }
            //endregion

            //region Set Encoding
            switch (properties.getEncoding()) {
                default:
                case Encoding_FM0:
                    setParamToReader(TMConstants.TMR_PARAM_GEN2_TAGENCODING, Gen2.TagEncoding.FM0);
                    break;
                case Encoding_M2:
                    setParamToReader(TMConstants.TMR_PARAM_GEN2_TAGENCODING, Gen2.TagEncoding.M2);
                    break;
                case Encoding_M4:
                    setParamToReader(TMConstants.TMR_PARAM_GEN2_TAGENCODING, Gen2.TagEncoding.M4);
                    break;
                case Encoding_M8:
                    setParamToReader(TMConstants.TMR_PARAM_GEN2_TAGENCODING, Gen2.TagEncoding.M8);
                    break;
            }
            //endregion

            //region Set Target
            switch (properties.getTarget()) {
                default:
                case Target_AB:
                    setParamToReader(TMConstants.TMR_PARAM_GEN2_TARGET, Gen2.Target.AB);
                    break;
                case Target_BA:
                    setParamToReader(TMConstants.TMR_PARAM_GEN2_TARGET, Gen2.Target.BA);
                    break;
                case Target_A:
                    setParamToReader(TMConstants.TMR_PARAM_GEN2_TARGET, Gen2.Target.A);
                    break;
                case Target_B:
                    setParamToReader(TMConstants.TMR_PARAM_GEN2_TARGET, Gen2.Target.B);
                    break;
            }
            //endregion

            //region Set Q
            switch (properties.getQValue()) {
                default:
                case 0:
                    setParamToReader(TMConstants.TMR_PARAM_GEN2_Q, new Gen2.DynamicQ());
                    break;
                case 1:
                case 2:
                case 3:
                case 4:
                case 5:
                case 6:
                case 7:
                case 8:
                case 9:
                case 10:
                case 11:
                case 12:
                case 13:
                case 14:
                case 15:
                case 16:
                    setParamToReader(TMConstants.TMR_PARAM_GEN2_Q, new Gen2.StaticQ(properties.getQValue() - 1));
                    break;
            }
            //endregion

            //region Set TARI value
            switch (properties.getTARI()) {
                default:
                case TARI_25:
                    setParamToReader(TMConstants.TMR_PARAM_GEN2_TARI, Gen2.Tari.TARI_25US);
                    break;
                case TARI_12_5:
                    setParamToReader(TMConstants.TMR_PARAM_GEN2_TARI, Gen2.Tari.TARI_12_5US);
                    break;
                case TARI_6_25:
                    setParamToReader(TMConstants.TMR_PARAM_GEN2_TARI, Gen2.Tari.TARI_6_25US);
                    break;
            }
            //endregion

            setParamToReader("/reader/read/asyncOnTime", onTime);
            setParamToReader("/reader/read/asyncOffTime", offTime);

            setParamToReader(TMConstants.TMR_PARAM_READER_STATUS_TEMPERATURE, true);

            setParamToReader(TMConstants.TMR_PARAM_TAGOP_ANTENNA, 1);

            SimpleReadPlan simplePlan;
            if (properties.getFastSearch()) {
                LoggerUtil.debug(TAG, "Fast search is On");
                simplePlan = new SimpleReadPlan(new int[]{1}, TagProtocol.GEN2, true);
            } else {
                LoggerUtil.debug(TAG, "Fast search is Off");
                simplePlan = new SimpleReadPlan(new int[]{1}, TagProtocol.GEN2, false);
            }
            reader.paramSet(TMConstants.TMR_PARAM_READ_PLAN, simplePlan);
        } catch (Exception ex) {
            LoggerUtil.error(TAG, "Exception : ", ex);
        }

    }

    public void setupReaderForLocate() {
        try {
            //region Set antenna power
            setParamToReader("/reader/radio/readPower", 2600);
            //endregion

            //region Set read timeout
            setParamToReader("/reader/commandTimeout", 1000);
            //endregion

            //region Set BLF
            setParamToReader(TMConstants.TMR_PARAM_GEN2_BLF, Gen2.LinkFrequency.LINK640KHZ);
            //endregion

            //region Set Session
            setParamToReader(TMConstants.TMR_PARAM_GEN2_SESSION, Gen2.Session.S0);
            //endregion

            //region Set Encoding
            setParamToReader(TMConstants.TMR_PARAM_GEN2_TAGENCODING, Gen2.TagEncoding.FM0);
            //endregion

            //region Set Target
            setParamToReader(TMConstants.TMR_PARAM_GEN2_TARGET, Gen2.Target.A);
            //endregion

            //region Set Q
            setParamToReader(TMConstants.TMR_PARAM_GEN2_Q, new Gen2.DynamicQ());
            //endregion

            //region Set TARI value
            setParamToReader(TMConstants.TMR_PARAM_GEN2_TARI, Gen2.Tari.TARI_6_25US);
            //endregion

            setParamToReader("/reader/read/asyncOnTime", 3000);
            setParamToReader("/reader/read/asyncOffTime", 300);

            setParamToReader(TMConstants.TMR_PARAM_READER_STATUS_TEMPERATURE, true);

            setParamToReader(TMConstants.TMR_PARAM_TAGOP_ANTENNA, 1);

            SimpleReadPlan simplePlan;
            simplePlan = new SimpleReadPlan(new int[]{1}, TagProtocol.GEN2, false);
            reader.paramSet(TMConstants.TMR_PARAM_READ_PLAN, simplePlan);
        } catch (Exception ex) {
            LoggerUtil.error(TAG, "Exception : ", ex);
        }

    }

    public void setTargetTag(String epc) throws ReaderException {
        SimpleReadPlan simplePlan;
        TagFilter filter = new TagData(epc);
        simplePlan = new SimpleReadPlan(new int[]{1}, TagProtocol.GEN2, filter, 100, false);
        reader.paramSet(TMConstants.TMR_PARAM_READ_PLAN, simplePlan);
    }

    public boolean setParamToReader(String key, Object value) {
        boolean result = false;
        int retryCount = 0;
        while (retryCount < 3) {
            retryCount++;
            try {
//                Object currVal = reader.paramGet(key);
//                if (currVal.toString().equalsIgnoreCase(value.toString())) {
//                    LoggerUtil.debug(TAG, String.format("Param %s already set as %s", key, currVal.toString()));
//                    return true;
//                }

                reader.paramSet(key, value);

                Object currVal = reader.paramGet(key);
                if (currVal.toString().equalsIgnoreCase(value.toString())) {
                    result = true;
                } else {
                    LoggerUtil.debug(TAG, String.format("Check param %s is %s", key, currVal.toString()));
                }

                if (result) {
                    break;
                } else {
                    Thread.sleep(200);
                }
            } catch (Exception e) {
                LoggerUtil.error(TAG, String.format("Exception in param %s", key), e);
            }
        }
        LoggerUtil.debug(TAG, String.format("Set param %s to %s : %s", key, value.toString(),
                result ? "successfully" : "failed"));
        return result;
    }

    public void updateRFIDVersion() {
        try {
            layout_firmware.setVisibility(View.VISIBLE);
            firmware_hint.setVisibility(View.GONE);

            String software = (String) reader.paramGet("/reader/version/software");
            softwareVersionView.setText(software);
            String hardware = (String) reader.paramGet("/reader/version/hardware");
            hardwareVersionView.setText(hardware);
            String model = (String) reader.paramGet("/reader/version/model");
            modelVersionView.setText(model);
            String productGroup = (String) reader.paramGet("/reader/version/productGroup");
            productGroupVersionView.setText(productGroup);
            String serial = (String) reader.paramGet("/reader/version/serial");
            serialNumberView.setText(serial);

            String baudrate = reader.paramGet(TMConstants.TMR_PARAM_BAUDRATE).toString();
            String temperature = reader.paramGet(TMConstants.TMR_PARAM_RADIO_TEMPERATURE).toString();

            LoggerUtil.debug(TAG, String.format("/reader/version/software = %s", software));
            LoggerUtil.debug(TAG, String.format("/reader/version/hardware = %s", hardware));
            LoggerUtil.debug(TAG, String.format("/reader/version/model = %s", model));
            LoggerUtil.debug(TAG, String.format("/reader/version/productGroup = %s", productGroup));
            LoggerUtil.debug(TAG, String.format("/reader/version/serial = %s", serial));

            LoggerUtil.debug(TAG, String.format("%s = %s", TMConstants.TMR_PARAM_BAUDRATE, baudrate));
            LoggerUtil.debug(TAG, String.format("%s = %s", TMConstants.TMR_PARAM_RADIO_TEMPERATURE, temperature));
        } catch (ReaderException e) {
            e.printStackTrace();
        }
    }
    //endregion
}