package com.unitech.rfidreader.action;

import android.content.BroadcastReceiver;
import android.content.Context;
import android.content.Intent;
import android.content.IntentFilter;
import android.os.Build;
import android.os.Bundle;
import android.text.InputFilter;
import android.text.InputType;
import android.text.Spanned;
import android.view.LayoutInflater;
import android.view.View;
import android.view.ViewGroup;
import android.widget.AdapterView;
import android.widget.ArrayAdapter;
import android.widget.Button;
import android.widget.EditText;
import android.widget.LinearLayout;
import android.widget.Spinner;
import android.widget.TextView;
import android.widget.Toast;

import com.thingmagic.Gen2;
import com.thingmagic.ReaderException;
import com.thingmagic.TMConstants;
import com.thingmagic.TagData;
import com.thingmagic.TagOp;
import com.unitech.rfidreader.R;
import com.unitech.rfidreader.ReaderActivity;
import com.unitech.util.LoggerUtil;

import androidx.annotation.NonNull;
import androidx.annotation.RequiresApi;
import androidx.fragment.app.Fragment;

/**
 * A placeholder fragment containing a simple view.
 */
public class PlaceholderFragment extends Fragment implements DataUpdateListener {

    private static final String TAG = "SectionsFragment";
    private Context mContext;
    private String mTabTitle;
    private ReaderActivity mReaderActivity;
    private Boolean startState = false;
    private HistogramData data = null;
    private int histogramSize = 20;
    private int histogramMax = -40;
    private int histogramMin = -80;
    private int histogramRange = histogramMax - histogramMin;


    public PlaceholderFragment(ReaderActivity readerActivity, String title) {
        mTabTitle = title.toUpperCase();
        mReaderActivity = readerActivity;
    }

    @Override
    public void onAttach(@NonNull Context context) {
        super.onAttach(context);
        mContext = context;
    }

    @Override
    public void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
    }

    @Override
    public View onCreateView(@NonNull LayoutInflater inflater, ViewGroup container, Bundle savedInstanceState) {
        View root = inflater.inflate(R.layout.fragment_action, container, false);
        findViewByID(root);
        TabView(mTabTitle);
        startState = false;
        IntentFilter filter = new IntentFilter();
        filter.addAction("ACTION_LOCATE_DATA");
        filter.addAction("ACTION_INITIAL");
        filter.addAction("ACTION_LOCATE_READING");
        mReaderActivity.registerReceiver(mBroadcastReceiver, filter);
        return root;
    }

    @Override
    public void onDestroyView() {
        mReaderActivity.unregisterReceiver(mBroadcastReceiver);
        super.onDestroyView();
    }

    @Override
    public void onDestroy() {
        super.onDestroy();
    }

    @Override
    public void onDetach() {
        super.onDetach();
    }


    private LinearLayout layout_epc;
    private LinearLayout layout_bank;
    private LinearLayout layout_password;
    private LinearLayout layout_address;
    private LinearLayout layout_length;
    private LinearLayout layout_data;
    private LinearLayout layout_privilege;
    private LinearLayout layout_histogram;
    private EditText edt_epc;
    private EditText edt_password;
    private EditText edt_address;
    private EditText edt_length;
    private EditText edt_data;
    private TextView txt_bank;
    private Spinner spinner_bank;
    private Spinner spinner_privilege;
    private Spinner spinner_DataType;
    private Button btn_execute;
    private Button btn_start;
    private HistogramView histogram;

    private void findViewByID(View root) {
        layout_epc = root.findViewById(R.id.layout_epc);
        layout_bank = root.findViewById(R.id.layout_bank);
        layout_password = root.findViewById(R.id.layout_password);
        layout_address = root.findViewById(R.id.layout_address);
        layout_length = root.findViewById(R.id.layout_length);
        layout_data = root.findViewById(R.id.layout_data);
        layout_privilege = root.findViewById(R.id.layout_privilege);
        layout_histogram = root.findViewById(R.id.layout_histogram);
        edt_epc = root.findViewById(R.id.edt_epc);
        edt_password = root.findViewById(R.id.edt_password);
        edt_address = root.findViewById(R.id.edt_address);
        edt_length = root.findViewById(R.id.edt_length);
        edt_data = root.findViewById(R.id.edt_data);
        spinner_bank = root.findViewById(R.id.spinner_bank);
        spinner_privilege = root.findViewById(R.id.spinner_privilege);
        spinner_DataType = root.findViewById(R.id.spinner_outputDataType);
        txt_bank = root.findViewById(R.id.txt_bank);
        btn_execute = root.findViewById(R.id.btn_execute);
        btn_start = root.findViewById(R.id.btn_start);
        histogram = root.findViewById(R.id.histogram);
        spinner_DataType.setOnItemSelectedListener(new AdapterView.OnItemSelectedListener() {
            @Override
            public void onItemSelected(AdapterView<?> parent, View view, int position, long id) {
                edt_data.setText("");
                InputFilter inputFilter_moodMsg;
                final char[] filter = new char[]{'0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'a', 'b', 'c', 'd', 'e', 'f', 'A', 'B', 'C', 'D', 'E', 'F'};
                switch (position) {
                    case 0:
                        inputFilter_moodMsg = new InputFilter() {
                            @Override
                            public CharSequence filter(CharSequence source, int start, int end, Spanned dest, int dstart, int dend) {
                                // Here you can add more controls, e.g. allow only hex chars etc
                                boolean getC = false;
                                for (int i = start; i < end; i++) {
                                    for (char c : filter) {
                                        if (source.charAt(i) == c) {
                                            getC = true;
                                        }
                                    }
                                }
                                return (getC) ? null : "";
                            }
                        };
                        edt_data.setFilters(new InputFilter[]{inputFilter_moodMsg});
                        edt_data.setInputType(InputType.TYPE_TEXT_FLAG_NO_SUGGESTIONS);
                        break;
                    case 1:
                        inputFilter_moodMsg = new InputFilter() {
                            @Override
                            public CharSequence filter(CharSequence source, int start, int end, Spanned dest, int dstart, int dend) {
                                // Here you can add more controls, e.g. allow only hex chars etc
                                return null;
                            }
                        };
                        edt_data.setFilters(new InputFilter[]{inputFilter_moodMsg});
                        edt_data.setInputType(InputType.TYPE_TEXT_FLAG_NO_SUGGESTIONS);
                        break;
                }
            }

            @Override
            public void onNothingSelected(AdapterView<?> parent) {

            }
        });
    }

    private void TabView(String TabTitle) {
//        int layout_resource = android.R.layout.simple_spinner_dropdown_item;
        int layout_resource = R.layout.spinner_item;
        switch (TabTitle) {
            case "READ":
            case "WRITE":
                layout_epc.setVisibility(View.VISIBLE);
                layout_bank.setVisibility(View.VISIBLE);
                layout_password.setVisibility(View.VISIBLE);
                layout_address.setVisibility(View.VISIBLE);
                layout_length.setVisibility(View.VISIBLE);
                layout_data.setVisibility(View.VISIBLE);
                layout_privilege.setVisibility(View.GONE);
                layout_histogram.setVisibility(View.GONE);
                btn_execute.setVisibility(View.VISIBLE);
                edt_password.setHint(R.string.access_password);
                spinner_DataType.setAdapter(new ArrayAdapter<>(mContext, layout_resource, mContext.getResources().getStringArray(R.array.data_output_type_array)));
//                spinner_passwordType.setAdapter(new ArrayAdapter<>(mContext, layout_resource, mContext.getResources().getStringArray(R.array.password_array)));
                spinner_bank.setAdapter(new ArrayAdapter<>(mContext, layout_resource, mContext.getResources().getStringArray(R.array.bank_array)));
//                spinner_passwordType.setSelection(0);//xxxxxxxx
                spinner_bank.setSelection(1);//EPC
                spinner_DataType.setSelection(0);//HEX
                edt_address.setText("2");
                edt_length.setText("6");
                edt_password.setText("00000000");
                break;
            case "LOCK":
                layout_epc.setVisibility(View.VISIBLE);
                layout_bank.setVisibility(View.VISIBLE);
                layout_password.setVisibility(View.VISIBLE);
                layout_address.setVisibility(View.GONE);
                layout_length.setVisibility(View.GONE);
                layout_data.setVisibility(View.GONE);
                layout_privilege.setVisibility(View.VISIBLE);
                layout_histogram.setVisibility(View.GONE);
                btn_execute.setVisibility(View.VISIBLE);
                edt_password.setHint(R.string.access_password);
                txt_bank.setText(R.string.field);
//                spinner_passwordType.setAdapter(new ArrayAdapter<>(mContext, layout_resource, mContext.getResources().getStringArray(R.array.password_array)));
                spinner_bank.setAdapter(new ArrayAdapter<>(mContext, layout_resource, mContext.getResources().getStringArray(R.array.bank_lock_array)));
                spinner_privilege.setAdapter(new ArrayAdapter<>(mContext, layout_resource, mContext.getResources().getStringArray(R.array.privilege_array)));
//                spinner_passwordType.setSelection(0);//xxxxxxxx
                spinner_bank.setSelection(2);//EPC
                spinner_privilege.setSelection(0);//LOCK
                edt_password.setText("00000000");
                break;
            case "KILL":
                layout_epc.setVisibility(View.VISIBLE);
                layout_bank.setVisibility(View.GONE);
                layout_password.setVisibility(View.VISIBLE);
                layout_address.setVisibility(View.GONE);
                layout_length.setVisibility(View.GONE);
                layout_data.setVisibility(View.GONE);
                layout_privilege.setVisibility(View.GONE);
                layout_histogram.setVisibility(View.GONE);
                btn_execute.setVisibility(View.VISIBLE);
                edt_password.setHint(R.string.kill_password);
//                spinner_passwordType.setAdapter(new ArrayAdapter<>(mContext, layout_resource, mContext.getResources().getStringArray(R.array.password_array)));
//                spinner_passwordType.setSelection(0);//xxxxxxxx
                edt_password.setText("00000000");
                break;
            case "LOCATE":
                layout_epc.setVisibility(View.VISIBLE);
                layout_bank.setVisibility(View.GONE);
                layout_password.setVisibility(View.GONE);
                layout_address.setVisibility(View.GONE);
                layout_length.setVisibility(View.GONE);
                layout_data.setVisibility(View.GONE);
                layout_privilege.setVisibility(View.GONE);
                layout_histogram.setVisibility(View.VISIBLE);
                btn_execute.setVisibility(View.GONE);
                data = new HistogramData(histogramSize);
                break;
        }
        //Change Button text
        if(edt_epc != null) {
            edt_epc.setText(mReaderActivity.targetEPC);
        }
        initialView();

        btn_execute.setText(TabTitle.equals("LOCK") ? "EXECUTE" : TabTitle);
        btn_execute.setOnClickListener(BtnEvent);
        btn_start.setOnClickListener(BtnEvent);
    }

    private String getPassword(String password) {
        return password.replaceAll(" ", "");
    }

    private boolean CheckPasswordType(String password) {
        int length = password.length();
        String[] spiltPassword;
        switch (length) {
            case 8:
                if (password.contains(" ")) {
                    return false;
                }
                break;
            case 9:
                if (!password.contains(" ")) {
                    return false;
                }
                spiltPassword = password.split(" ");
                if (spiltPassword.length != 2) {
                    return false;
                }
                for (String s : spiltPassword) {
                    if (s.length() != (8 / spiltPassword.length)) {
                        return false;
                    }
                }
                break;
            case 11:
                if (!password.contains(" ")) {
                    return false;
                }
                spiltPassword = password.split(" ");
                if (spiltPassword.length != 4) {
                    return false;
                }
                for (String s : spiltPassword) {
                    if (s.length() != (8 / spiltPassword.length)) {
                        return false;
                    }
                }
                break;
            default:
                return false;
        }
        return true;
    }


    private View.OnClickListener BtnEvent = new View.OnClickListener() {
        @RequiresApi(api = Build.VERSION_CODES.O)
        @Override
        public void onClick(View v) {
            String targetTag = getTargetEPC();

            if (targetTag.trim().equalsIgnoreCase(""))
            {
                Toast.makeText(mReaderActivity, "Please input target tag!", Toast.LENGTH_SHORT).show();
                return;
            }

            if (mReaderActivity.serviceListener.isRunning() && !mTabTitle.equals("LOCATE"))
                return;

            switch (mTabTitle) {
                case "READ":
                    try {
                        edt_data.setText("");
                        String strEPC = edt_epc.getText().toString().trim();
                        String password = edt_password.getText().toString().trim();
                        String address = edt_address.getText().toString().trim();
                        String addressLength = edt_length.getText().toString().trim();

                        password = password.equals("") ? "00000000" : password;
                        if (!CheckPasswordType(password)) {
                            Toast.makeText(mReaderActivity, "Please input correct password type.", Toast.LENGTH_SHORT).show();
                            break;
                        }
                        password = getPassword(password);
                        int iPass = Integer.parseUnsignedInt(password, 16);
                        if (address.equals("") || addressLength.equals("")) {
                            Toast.makeText(mReaderActivity, "Please input data !!!", Toast.LENGTH_SHORT).show();
                            break;
                        }
                        int offset = Integer.parseInt(address);
                        byte length = Integer.valueOf(addressLength).byteValue();
                        if (offset < 0 || length <= 0) {
                            Toast.makeText(mReaderActivity, "Please input correct data !!!", Toast.LENGTH_SHORT).show();
                            break;
                        }

                        Gen2.Password PW = new Gen2.Password(iPass);
                        mReaderActivity.setParamToReader(TMConstants.TMR_PARAM_GEN2_ACCESSPASSWORD, PW);

//                        mReaderActivity.setupReaderForLocate();
//                        mReaderActivity.setupReader(2000, 600);
                        Gen2.ReadData tagRead = new Gen2.ReadData(Gen2.Bank.getBank(getBank()), offset, length);
                        long startTime = NowTime();
                        short[] data = (short[]) mReaderActivity.reader.executeTagOp(tagRead, new TagData(strEPC));
                        long duration = NowTime() - startTime;
                        StringBuilder str = new StringBuilder();
                        for (short aData : data) {
                            String temp = Integer.toHexString(aData & 0xffff);
                            while (temp.length() < 4) {
                                StringBuilder sb = new StringBuilder();
                                temp = sb.append("0").append(temp).toString();
                            }
                            str.append(temp);
                        }
                        String HexString = str.toString();
                        if (spinner_DataType.getSelectedItem().toString().equals("ASCII")) {
                            StringBuilder outputString = new StringBuilder();
                            for (int i = 0; i < HexString.length(); i += 2) {
                                String temp = HexString.substring(i, i + 2);
                                outputString.append((char) Integer.parseInt(temp, 16));
                            }
                            edt_data.setText(outputString.toString());
                        } else {
                            edt_data.setText(HexString.toUpperCase());
                        }

                        String msg = "Read success in " + duration + " ms";
                        Toast.makeText(mReaderActivity, msg, Toast.LENGTH_SHORT).show();
                        LoggerUtil.info(TAG, msg);

                    } catch (Exception ex) {
                        LoggerUtil.error(TAG, "Read Exception", ex);
                        Toast.makeText(mReaderActivity, "Read fail.\n" + ex.getMessage(), Toast.LENGTH_SHORT).show();
                    }
                    break;
                case "WRITE":
                    try {
                        String strEPC = edt_epc.getText().toString().trim();
                        String password = edt_password.getText().toString().trim();
                        String address = edt_address.getText().toString().trim();
                        String addressLength = edt_length.getText().toString().trim();
                        String strData = edt_data.getText().toString().trim().toLowerCase();

                        password = password.equals("") ? "00000000" : password;
                        if (!CheckPasswordType(password)) {
                            Toast.makeText(mReaderActivity, "Please input correct password type.", Toast.LENGTH_SHORT).show();
                            break;
                        }
                        password = getPassword(password);
                        int iPass = Integer.parseUnsignedInt(password, 16);
                        if (address.equals("") || addressLength.equals("")) {
                            Toast.makeText(mReaderActivity, "Please input data !!!", Toast.LENGTH_SHORT).show();
                            break;
                        }
                        int offset = Integer.parseInt(address);
                        byte length = Integer.valueOf(addressLength).byteValue();
                        if (offset < 0 || length <= 0) {
                            Toast.makeText(mReaderActivity, "Please input correct data !!!", Toast.LENGTH_SHORT).show();
                            break;
                        }

                        if (strData.length() != length * 4 && spinner_DataType.getSelectedItem().toString().equals("HEX")) {
                            Toast.makeText(mReaderActivity, "Please input a " + length * 4 + " lengths data", Toast.LENGTH_SHORT).show();
                            break;
                        }
                        if (strData.length() != length * 2 && spinner_DataType.getSelectedItem().toString().equals("ASCII")) {
                            Toast.makeText(mReaderActivity, "Please input a " + length * 2 + " lengths data", Toast.LENGTH_SHORT).show();
                            break;
                        }

//                        mReaderActivity.setupReaderForLocate();
//                        mReaderActivity.setupReader(3500, 600);
                        Gen2.Password PW = new Gen2.Password(iPass);
                        mReaderActivity.setParamToReader(TMConstants.TMR_PARAM_GEN2_ACCESSPASSWORD, PW);

                        if (spinner_DataType.getSelectedItem().toString().equals("ASCII")) {
                            strData = toASCII(strData);
                        }
                        Gen2.WriteData tagWrite = new Gen2.WriteData(Gen2.Bank.getBank(getBank()), offset, toShorts(strData));
                        long startTime = NowTime();
                        TagOp tagOp = (TagOp) mReaderActivity.reader.executeTagOp(tagWrite, new TagData(strEPC));
                        long duration = NowTime() - startTime;
                        String msg = "Write success in " + duration + " ms";
                        Toast.makeText(mReaderActivity, msg, Toast.LENGTH_SHORT).show();
                        LoggerUtil.info(TAG, msg);

                    } catch (Exception ex) {
                        LoggerUtil.error(TAG, "Write Exception", ex);
                        Toast.makeText(mReaderActivity, "WRITE fail.\n" + ex.getMessage(), Toast.LENGTH_SHORT).show();
                    }

                    break;
                case "LOCK":
                    String[] spinner_select_item_text = spinner_bank.getSelectedItem().toString().toUpperCase().split(" ");
                    String checkLockAction = spinner_select_item_text[0] + "_" + spinner_privilege.getSelectedItem().toString().toUpperCase();
                    try {
                        String password = edt_password.getText().toString().trim();
                        password = password.equals("") ? "00000000" : password;
                        if (!CheckPasswordType(password)) {
                            Toast.makeText(mReaderActivity, "Please input correct password type.", Toast.LENGTH_SHORT).show();
                            break;
                        }
                        password = getPassword(password);
                        int iPass = Integer.parseUnsignedInt(password, 16);
                        if (iPass == 0) {
                            Toast.makeText(mReaderActivity, "Please set your access password not be zero.", Toast.LENGTH_SHORT).show();
                            break;
                        }
//                        mReaderActivity.setupReaderForLocate();
//                        mReaderActivity.setupReader(3500, 600);
                        Gen2.Password PW = new Gen2.Password(iPass);
                        mReaderActivity.setParamToReader(TMConstants.TMR_PARAM_GEN2_ACCESSPASSWORD, PW);

                        Gen2.LockAction lockAction = Gen2.LockAction.parse(checkLockAction);
                        Gen2.Lock tagLock = new Gen2.Lock(iPass, lockAction);
                        TagData target = new TagData(edt_epc.getText().toString());
                        long startTime = NowTime();
                        TagOp tagOp = (TagOp) mReaderActivity.reader.executeTagOp(tagLock, target);
                        long duration = NowTime() - startTime;
                        String msg = checkLockAction + " success in " + duration + " ms";
                        Toast.makeText(mReaderActivity, msg, Toast.LENGTH_SHORT).show();
                        LoggerUtil.info(TAG, msg);

                    } catch (Exception ex) {
                        LoggerUtil.error(TAG, "Lock Exception", ex);
                        Toast.makeText(mReaderActivity, checkLockAction + " fail.\n" + ex.getMessage(), Toast.LENGTH_SHORT).show();
                    }
                    break;
                case "KILL":
                    try {
                        String password = edt_password.getText().toString().trim();
                        password = password.equals("") ? "00000000" : password;
                        if (!CheckPasswordType(password)) {
                            Toast.makeText(mReaderActivity, "Please input correct password type.", Toast.LENGTH_SHORT).show();
                            break;
                        }
                        password = getPassword(password);

                        int iPass = Integer.parseUnsignedInt(password, 16);
                        if (iPass == 0) {
                            Toast.makeText(mReaderActivity, "Please set your kill password not be zero.", Toast.LENGTH_SHORT).show();
                            break;
                        }
//                        mReaderActivity.setupReaderForLocate();
//                        mReaderActivity.setupReader(3500, 600);
//                        mReaderActivity.setParamToReader(TMConstants.TMR_PARAM_TAGOP_ANTENNA, 1);

                        TagData target = new TagData(edt_epc.getText().toString());
                        Gen2.Kill tagKill = new Gen2.Kill(iPass);
                        long startTime = System.currentTimeMillis();
                        TagOp tagOp = (TagOp) mReaderActivity.reader.executeTagOp(tagKill, target);
                        long duration = System.currentTimeMillis() - startTime;
                        String msg = "Kill tag success in " + duration + " ms.";
                        Toast.makeText(mReaderActivity, msg, Toast.LENGTH_SHORT).show();

                    } catch (Exception ex) {
                        LoggerUtil.error(TAG, "Kill Exception", ex);
                        Toast.makeText(mReaderActivity, "Kill tag fail.\n" + ex.getMessage(), Toast.LENGTH_SHORT).show();
                    }
                    break;
                case "LOCATE":
                    setStartState(!startState, 1);
                    break;

            }
        }
    };


    private long NowTime() {
        return System.currentTimeMillis();
    }

    private int targetRSSI = -90;
    private long preBeepTime = 0;

    private void UpdateHistogram() {
        int size = (targetRSSI - histogramMax) / (histogramRange / histogramSize) + histogramSize;
        size = Math.min(size, histogramSize);
        long nowTime = NowTime();
        if (size > 0) {
            if (preBeepTime != 0) {
                if (targetRSSI > -40)
                {
                    preBeepTime = nowTime;
                    mReaderActivity.mBeeper.beepLocate(size, histogramSize);
                }
                else if (targetRSSI > -50)
                {
                    if (nowTime - preBeepTime > 100) {
                        preBeepTime = nowTime;
                        mReaderActivity.mBeeper.beepLocate(size, histogramSize);
                    }
                }
                else if (targetRSSI > -60)
                {
                    if (nowTime - preBeepTime > 300) {
                        preBeepTime = nowTime;
                        mReaderActivity.mBeeper.beepLocate(size, histogramSize);
                    }
                }
                else if (targetRSSI > -70)
                {
                    if (nowTime - preBeepTime > 500) {
                        preBeepTime = nowTime;
                        mReaderActivity.mBeeper.beepLocate(size, histogramSize);
                    }
                }
                else
                {
                    if (nowTime - preBeepTime > 800) {
                        preBeepTime = nowTime;
                        mReaderActivity.mBeeper.beepLocate(size, histogramSize);
                    }
                }
            }
            else {
                preBeepTime = nowTime;
                mReaderActivity.mBeeper.beepLocate(size, histogramSize);
            }

        }
        data.setData(size, targetRSSI + "dbm");

        histogram.setData(data);
    }

    private BroadcastReceiver mBroadcastReceiver = new BroadcastReceiver() {
        @Override
        public void onReceive(Context context, Intent intent) {
            String action = intent.getAction();
            switch (action) {
                case "ACTION_LOCATE_DATA":
                    if (mTabTitle.equals("LOCATE") && startState) {
                        String epc = intent.getStringExtra("LOCATE_EPC");
                        int rssi = intent.getIntExtra("LOCATE_RSSI", histogramMin);
                        targetRSSI = rssi;
                        UpdateHistogram();
                    }
                    break;
                case "ACTION_INITIAL":
                    initialView();
                    break;

                case "ACTION_LOCATE_READING":
                    if (mTabTitle.equals("LOCATE")) {
                        boolean startStatus = intent.getBooleanExtra("start", false);
                        int startEnable = intent.getIntExtra("enable", 1);
                        if (!startStatus || startEnable != 0) {
                            setStartState(startStatus, startEnable);
                        }
                    }
                    break;
            }
        }
    };

    private int getBank() {
        int temp;
        switch (spinner_bank.getSelectedItem().toString()) {
            case "RESERVED":
                temp = 0;
                break;
            case "EPC":
                temp = 1;
                break;
            case "TID":
                temp = 2;
                break;
            case "USER":
                temp = 3;
                break;
            default:
                temp = -1;
                break;
        }
        return temp;
    }

    private String toASCII(String input) {
        StringBuilder str = new StringBuilder();
        for (int i = 0; i < input.length(); i++) {
            char c = input.charAt(i);
            str.append(Integer.toHexString(c));
        }
        return str.toString();
    }

    private short[] toShorts(String finalString) {
        StringBuilder ChangeStr = new StringBuilder();
        for (int i = 0; i < finalString.length(); i++) {
            ChangeStr.append(finalString.charAt(i));
            if (i % 4 == 3) {
                ChangeStr.append(" ");
            }
        }
        String[] tabOfShortString = ChangeStr.toString().split(" ");
        int length = tabOfShortString.length;
        short[] shortsArray = new short[length];
        for (int i = 0; i < length; i++) {
            short res = (short) Integer.parseInt(tabOfShortString[i], 16);
            shortsArray[i] = res;
        }

        return shortsArray;
    }

    private void initialView() {
        if (edt_data != null) {
            edt_data.setText("");
        }
        setStartState(false, 1);
    }

    private void setStartState(boolean state, int enable) {
        startState = state;
        if (btn_start == null) {
            return;
        }
        if (state) {
            btn_start.setBackground(mReaderActivity.getDrawable(R.drawable.stop));
        } else {
            btn_start.setBackground(mReaderActivity.getDrawable(R.drawable.start));
        }
        switch (enable) {
            case 0:
                btn_start.setEnabled(false);
                btn_start.setClickable(false);
                break;
            case 1:
                btn_start.setEnabled(true);
                btn_start.setClickable(true);
                break;
        }

        if (data == null) {
            return;
        }
        data.setData(0, "0");
        histogram.setData(data);
        if (mReaderActivity.getReadBtnText().equals(getString(R.string.start_reading)) &&
                mReaderActivity.reader != null &&
                mReaderActivity.getConnectBtnText().equals(getString(R.string.Disconnect)) &&
                mReaderActivity.serviceListener != null &&
                startState) {
            //Harvey should setup the Inventory parameter for locate
            try {
                mReaderActivity.setTargetTag(edt_epc.getText().toString());
            } catch (ReaderException e) {
                e.printStackTrace();
            }

            mReaderActivity.serviceListener.startReading();
        }
        if (mReaderActivity.getReadBtnText().equals(getString(R.string.stop_reading)) &&
                mReaderActivity.reader != null &&
                mReaderActivity.getConnectBtnText().equals(getString(R.string.Disconnect)) &&
                mReaderActivity.serviceListener != null &&
                !startState) {
            mReaderActivity.serviceListener.stopReading();
        }

    }

    @Override
    public void onDataUpdate(String mEPC) {
        if (edt_epc != null) {
            edt_epc.setText(mEPC);
        }

        initialView();
    }

    public String getTargetEPC()
    {
        if (edt_epc != null) {
            return edt_epc.getText().toString();
        }
        return "";
    }

    public String getTitle()
    {
        return mTabTitle;
    }

}